#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import warnings
from functools import reduce
import re
from .file_download import update_index
from .file_tools import validate_version, get_version_files_paths
from .dataframe_tools import add_index_levels, join_col_to_dataframe, sort_df_by_sample_status
from .exceptions import *

import cptac.utils as ut

class Dataset:
    """
    Note that all cancer datasets are class objects that inherit from cptac.dataset. Therefore
    the same function calls exist for cptac.Endometrial, cptac.Colon, etc.
    """

    def __init__(self, cancer_type, version, valid_versions, data_files, no_internet, attempt_update_index=True, skip_init=False):
        """Initialize variables for a Dataset object.

        Parameters:
        cancer_type (str): The cancer type requested for this dataset
        version (str): The version number requested for this dataset
        valid_versions (list of str): A list of all possible valid versions for this dataset
        data_files (dict, keys of str, values of list of str): A dictionary where the keys are the existing version of the dataset, and the values are lists of the data file names for that version.
        """
         # Initialize dataframe and definitions dicts as empty for this parent class
        self._data = {}
        self._helper_tables = {}
        self._definitions = {}
        self._readme_files = {}

         # Assign the valid dfs lists, but make them instance variables so they're easy to override if needed
         # These are the omics dataframes that are valid for use in the utilities functions
        self._valid_omics_dfs = [
            'acetylproteomics',
            'circular_RNA',
            'CNV',
            'lincRNA',
            'lipidomics',
            'metabolomics',
            'miRNA',
            'phosphoproteomics',
            'phosphoproteomics_gene',
            'proteomics',
            'somatic_mutation_binary',
            'transcriptomics', 
            'CNV_log2ratio',
            'CNV_gistic'
            
             ]

         # These are the metadata dataframes that are valid for use in the utilities functions
        self._valid_metadata_dfs = [
            "clinical",
            "derived_molecular",
            "experimental_design",
            #"followup", # Right now there are duplicate rows, so don't include follow up tables for joins.
            ] # We don't allow the treatment df, as in Ovarian, or medical_history df, as in Ccrcc, because they both have multiple rows for each sample.
        
        # Initialize the _cancer_type instance variable
        self._cancer_type = cancer_type.lower()
        
        if not skip_init:

           

            # Update the index, if possible and desired.
            if attempt_update_index and not no_internet:
                try:
                    update_index(self._cancer_type)
                except NoInternetError:
                    pass

            # Validate the version
            self._version = validate_version(version, self._cancer_type, use_context="init", valid_versions=valid_versions)

            # Get the paths to the data files
            version_data_files = data_files[self._version] # Get the data files for this version from the data files dictionary
            self._data_files_paths = get_version_files_paths(self._cancer_type, self._version, version_data_files)

           

    # Methods to get metadata dataframes
    def get_clinical(self, tissue_type="both"):
        """Get the clinical dataframe."""
        return self._get_dataframe("clinical", tissue_type)

    def get_derived_molecular(self, tissue_type="both"):
        """Get the derived_molecular dataframe."""
        return self._get_dataframe("derived_molecular", tissue_type)

    def get_experimental_design(self, tissue_type="both"):
        """Get the experimental_design dataframe."""
        return self._get_dataframe("experimental_design", tissue_type)

    def get_medical_history(self, tissue_type="both"):
        """Get the medical_history dataframe."""
        return self._get_dataframe("medical_history", tissue_type)

    def get_treatment(self, tissue_type="both"):
        """Get the treatment dataframe."""
        return self._get_dataframe("treatment", tissue_type)

    def get_followup(self, tissue_type="both"):
        """Get the followup dataframe."""
        return self._get_dataframe("followup", tissue_type)

    # Methods to get omics dataframes
    def get_acetylproteomics(self, tissue_type="both"):
        """Get the acetylproteomics dataframe."""
        return self._get_dataframe("acetylproteomics", tissue_type)
    
    def get_acetylproteomics_gene(self, tissue_type="both"):
        """Get the acetylproteomics_gene dataframe. The gene level acetylation measurement is an aggregate metric which potentially averages together individual measurements of different sites. Use get_acetylproteomics() to view the data for individual sites."""
        return self._get_dataframe("acetylproteomics_gene",tissue_type)

    def get_circular_RNA(self, tissue_type="both"):
        """Get the circular_RNA dataframe."""
        return self._get_dataframe("circular_RNA", tissue_type)

    def get_CNV(self):
        """Get the CNV dataframe."""
        return self._get_dataframe("CNV")

    def get_lincRNA(self, tissue_type="both"):
        """Get the lincRNA dataframe."""
        return self._get_dataframe("lincRNA",tissue_type)

    def get_lipidomics(self, tissue_type="both"):
        """Get the lipidomics dataframe."""
        return self._get_dataframe("lipidomics", tissue_type)

    def get_metabolomics(self, tissue_type="both"):
        """Get the metabolomics dataframe."""
        return self._get_dataframe("metabolomics", tissue_type)

    def get_methylation(self, tissue_type="both"):
        """Get the methylation dataframe."""
        return self._get_dataframe("methylation", tissue_type)

    def get_miRNA(self, tissue_type="both"):
        """Get the miRNA dataframe."""
        return self._get_dataframe("miRNA",tissue_type)

    def get_phosphoproteomics(self, tissue_type="both"):
        """Get the phosphoproteomics dataframe."""
        return self._get_dataframe("phosphoproteomics",tissue_type)

    def get_phosphoproteomics_gene(self, tissue_type="both"):
        """Get the phosphoproteomics_gene dataframe. The gene level phosphorylation measurement is an aggregate metric which potentially averages together individual measurements of different sites. Use get_phosphoproteomics() to view the data for individual sites."""
        return self._get_dataframe("phosphoproteomics_gene",tissue_type)
    
    def get_targeted_phosphoproteomics(self, tissue_type="both"):
        """Get the targeted_phosphoproteomics dataframe."""
        return self._get_dataframe("targeted_phosphoproteomics",tissue_type)

    def get_phosphosites(self, genes):
        """Returns dataframe with all phosphosites of specified gene or list of genes.

        Parameters:
        genes (str, or list or array-like of str): gene or list of genes to use to select phosphosites. str if single, list or array-like of str if multiple.

        Returns:
        pandas.DataFrame: The phosphoproteomics for the specified gene(s).
        """
        return self._get_omics_cols("phosphoproteomics", genes)

    def get_proteomics(self, tissue_type="both"):
        """Get the proteomics dataframe."""
        return self._get_dataframe("proteomics",tissue_type)
    
    def get_targeted_proteomics(self, tissue_type="both"):
        """Get the targeted_proteomics dataframe."""
        return self._get_dataframe("targeted_proteomics",tissue_type)

    def get_transcriptomics(self, tissue_type="both"):
        """Get the transcriptomics dataframe."""
        return self._get_dataframe("transcriptomics", tissue_type)

    def get_ubiquitinomics(self, tissue_type="both"):
        """Get the ubiquitinomics dataframe."""
        return self._get_dataframe("ubiquitinomics", tissue_type)

    # Methods to get mutations dataframes
    def get_gene_fusion(self):
        """Get the gene_fusion dataframe."""
        return self._get_dataframe("gene_fusion")

    def get_somatic_mutation(self):
        """Get the somatic_mutation dataframe."""
        return self._get_dataframe("somatic_mutation")

    def get_somatic_mutation_binary(self):
        """Get the somatic_mutation_binary dataframe, which has a binary value indicating, for each location on each gene, whether there was a mutation in that gene at that location, for each sample."""
        return self._get_dataframe("somatic_mutation_binary")

    # Help methods
    def define(self, term):
        """Print the definition a term, if it is in the dataset's list of definitions.

        Parameters:
        term (str): term to be defined

        Returns: None
        """
        if len(self._definitions.keys()) == 0:
            raise NoDefinitionsError("No definitions provided for this dataset.")
        elif term in self._definitions.keys():
            print(self._definitions[term])
        else:
            raise InvalidParameterError("{} not found in definitions. Check capitalization. Alternatively, the 'search(<your term>)' function, available for import from the cptac.utils sub-module, can be used to perform a web search of the term provided.".format(term))

    def get_cancer_type(self):
        """Return the cancer type for this dataset, as a string."""
        return self._cancer_type

    def version(self):
        """Return the dataset version of this instance, as a string."""
        return self._version

    def how_to_cite(self, cancer_type='', pmid='', unpublished=False):
        """Print instructions for citing the data."""
        ### 
        # OLD message: 
        # print("For use of the cptac Python data API, please cite 
        # our publication: https://www.biorxiv.org/content/10.1101/2020.11.16.385427v1")
        ###

        # current main message
        main_message = ('Please include the following statement(s) in publications using data '
         'accessed through this module:\n"Data used in this publication were generated '
         'by the Clinical Proteomic Tumor Analysis Consortium (NCI/NIH). '
         'Data were accessed through the Python module cptac, PMID: 33560848."')
        print(main_message)

        # extra message specific to cancer type
        if (cancer_type and pmid):
            cancer_specific_message = (f"Data from {cancer_type} were originally published in PMID: {pmid}")
            print(cancer_specific_message)
        elif (cancer_type and unpublished):
            print(f"Data for {cancer_type} has not been published yet.")
        else:
            # no additional message will be printed if we have not passed in parameters
            pass

    def get_data_list(self):
        data_list = {}
        for name in sorted(self._data.keys(), key=str.lower):
            df = self._data[name]
            data_list[name] = {'rows': df.shape[0], 'columns': df.shape[1]}
        return data_list

    def list_data(self):
        """Print list of loaded dataframes and dimensions."""
        print("Below are the dataframes contained in this dataset and their dimensions:\n")
        datasets = self.get_data_list()
        for entry in datasets:
            data = datasets[entry]
            print(f"{entry}")
            print(f"\t{data['rows']} rows")
            print(f"\t{data['columns']} columns")

    def list_definitions(self):
        """Print all terms defined in the dataset's list of definitions."""
        if len(self._definitions.keys()) > 0:
            for term in sorted(self._definitions.keys(), key=str.lower):
                print(term)
        else:
            raise NoDefinitionsError("No definitions provided for this dataset.")

    def get_genotype_all_vars(self, mutations_genes, mutations_filter=None, show_location=True, mutation_hotspot=None):
        """Return a dataframe that has the mutation type and whether or not it is a multiple mutation
        Parameters:
        mutation_genes (str, or list or array-like of str): The gene(s) to get mutation data for.
        mutations_filter (list, optional):  List of mutations to prioritize when filtering out multiple mutations, in order of priority.
        show_location (bool, optional): Whether to include the Location column from the mutation dataframe. Defaults to True.
        mutation_hotspot (optional): a list of hotspots
        """

        #If they don't give us a filter, this is the default.

        if mutations_filter == None:
            if self.get_cancer_type() == "colon":
                mutations_filter = ["Deletion", #deletion
                                        'frameshift deletion', 'frameshift insertion', 'frameshift substitution', 'stopgain', 'stoploss', #truncation
                                        'Missense_Mutation_hotspot',
    	                                'nonframeshift deletion', 'nonframeshift insertion', 'nonframeshift substitution', 'nonsynonymous SNV', #missense
                                        'Amplification',
                                         'Wildtype']


            elif self.get_cancer_type() == "hnscc":
                mutations_filter = ["Deletion", #deletion
                                        'Frame_Shift_Del', 'Frame_Shift_Ins', 'Nonsense_Mutation', 'Nonstop_Mutation', #truncation
                                        'Missense_Mutation_hotspot',
    	                                'Missense_Mutation',
                                        'Amplification',
                                        'In_Frame_Del', 'In_Frame_Ins', 'Splice_Site', #inframe changes
                                        'Silent','Wildtype']

            elif self.get_cancer_type() == "gbm":
                mutations_filter = ["Deletion", #deletion
                                        'Frame_Shift_Del', 'Frame_Shift_Ins', 'Nonsense_Mutation', 'Nonstop_Mutation', #truncation
                                        'Missense_Mutation_hotspot',
                                        'Missense_Mutation',
                                        'Amplification',
                                        'In_Frame_Del', 'In_Frame_Ins', 'Splice_Site', #inframe changes
                                        'Silent','Wildtype']

            else:
                mutations_filter = ["Deletion",
                                        'Frame_Shift_Del', 'Frame_Shift_Ins', 'Nonsense_Mutation', 'Nonstop_Mutation', #tuncation
                                        'Missense_Mutation_hotspot',
    	                                'Missense_Mutation',
                                        'Amplification',
                                        'In_Frame_Del', 'In_Frame_Ins', 'Splice_Site',
                                        'Silent',
                                        'Wildtype']

        if self.get_cancer_type() == 'colon':
            truncations = ['frameshift deletion', 'frameshift insertion', 'frameshift substitution', 'stopgain', 'stoploss']
            missenses = ['nonframeshift deletion', 'nonframeshift insertion', 'nonframeshift substitution', 'nonsynonymous SNV']
        elif self.get_cancer_type() == 'hnscc' and self.version() == "0.1":
            truncations =["stopgain", "stoploss"]
            missenses = ["nonframeshift insertion", "nonframeshift deletion"]
        else:
            truncations = ['Frame_Shift_Del', 'Frame_Shift_Ins', 'Nonsense_Mutation', 'Nonstop_Mutation', 'Splice_Site']
            missenses = ['In_Frame_Del', 'In_Frame_Ins', 'Missense_Mutation']

        if self.get_cancer_type() == "gbm":
            noncodings = ["Intron", "RNA", "3'Flank", "Splice_Region", "5'UTR", "5'Flank", "3'UTR"]



        #check that gene is in the somatic_mutation DataFrame
        somatic_mutation = self.get_somatic_mutation()
        if mutations_genes not in somatic_mutation["Gene"].unique(): #if the gene isn't in the somacic mutations df it will still have CNV data that we want
            def add_del_and_amp_no_somatic(row):
                if row[mutations_genes] <= -.2:
                    mutations = 'Deletion'

                elif row[mutations_genes] >= .2:
                    mutations = 'Amplification'
                else:
                    mutations = "No_Mutation"

                return mutations
            
            if self.get_cancer_type() == "ucecconf":
                cnv = self.get_CNV("log2ratio")
            else:      
                cnv = self.get_CNV()
                
            #drop the database index from ccrcc and brca
            if isinstance(cnv.keys(), pd.core.indexes.multi.MultiIndex):
                drop = ['Database_ID']
                cnv = ut.reduce_multiindex(df=cnv, levels_to_drop=drop)       
            gene_cnv = cnv[[mutations_genes]]
            mutation_col = gene_cnv.apply(add_del_and_amp_no_somatic, axis=1)
            df = gene_cnv.assign(Mutation = mutation_col)
            return df
 #combine the cnv and mutations dataframe
        if self.get_cancer_type() == "ucecconf":
            combined = self.join_omics_to_mutations(omics_df_name="CNV_log2ratio", mutations_genes=mutations_genes, omics_genes=mutations_genes)
            combined.columns = combined.columns.str.rstrip('_log2ratio') # strip log2ratio 


        else:
            combined = self.join_omics_to_mutations(omics_df_name="CNV", mutations_genes=mutations_genes, omics_genes=mutations_genes)


        #drop the database index from ccrcc
        if self.get_cancer_type() == "ccrcc" or self.get_cancer_type() == "brca":
             cc = self.get_CNV()#not sure about this step
             drop = ['Database_ID']
             combined = ut.reduce_multiindex(df=combined, levels_to_drop=drop)


        #If there are hotspot mutations, append 'hotspot' to the mutation type so that it's prioritized correctly
        def mark_hotspot_locations(row):
            #iterate through each location in the current row
            mutations = []
            for location in row[mutations_genes+'_Location']:
                if location in mutation_hotspot: #if it's a hotspot mutation
                    #get the position of the location
                    position = row[mutations_genes+'_Location'].index(location)
                    #use that to change the correct mutation
                    mutations.append(row[mutations_genes+"_Mutation"][position] + "_hotspot")
                else:
                    # get the position of the location
                    position = row[mutations_genes+'_Location'].index(location)
                    mutations.append(row[mutations_genes+"_Mutation"][position])
            return mutations

        if mutation_hotspot is not None:
            combined['hotspot'] = combined.apply(mark_hotspot_locations, axis=1)
            combined[mutations_genes+"_Mutation"] = combined['hotspot']
            combined = combined.drop(columns='hotspot')

        # Based on cnv make a new column with mutation type that includes deletions and amplifications
        def add_del_and_amp(row):
            if row[mutations_genes+"_CNV"] <= -.2:
                mutations = row[mutations_genes+"_Mutation"] + ['Deletion']
                locations = row[mutations_genes+'_Location']+['Deletion']

            elif row[mutations_genes+"_CNV"] >= .2:
                mutations = row[mutations_genes+"_Mutation"] + ['Amplification']
                locations = row[mutations_genes+'_Location']+['Amplification']
            else:
                mutations = row[mutations_genes+"_Mutation"]
                locations = row[mutations_genes+"_Location"]

            return mutations, locations


        combined['mutations'], combined['locations'] = zip(*combined.apply(add_del_and_amp, axis=1))


        #now that we have the deletion and amplifications, we need to prioritize the correct mutations.
        def sort(row):
            sortedcol = []
            location = []
            chosen_indices = []
            sample_mutations_list = row['mutations']
            sample_locations_list = row['locations']
            if len(sample_mutations_list) == 1: #if there's only one mutation in the list
                sortedcol.append(sample_mutations_list[0])
                location.append(sample_locations_list[0])

            else:
             
                for filter_val in mutations_filter: # This will start at the beginning of the filter list, thus filters earlier in the list are prioritized, like we want
                    if filter_val in sample_mutations_list:
                        chosen_indices = [index for index, value in enumerate(sample_mutations_list) if value == filter_val]
                    if len(chosen_indices) > 0: # We found at least one mutation from the filter to prioritize, so we don't need to worry about later values in the filter priority list
                        break

                if len(chosen_indices) == 0: # None of the mutations for the sample were in the filter, so we're going to have to use our default hierarchy
                    for mutation in sample_mutations_list:
                        if mutation in truncations:
                            chosen_indices += [index for index, value in enumerate(sample_mutations_list) if value == mutation]

                if len(chosen_indices) == 0: # None of them were in the filter, nor were truncations, so we'll grab all the missenses
                    for mutation in sample_mutations_list:
                        if mutation in missenses:
                            chosen_indices += [index for index, value in enumerate(sample_mutations_list) if value == mutation]

                if self.get_cancer_type() == "gbm" and len(chosen_indices) == 0: # None of them were in the filter, nor were truncations, nor missenses, so we'll grab all the noncodings
                    for mutation in sample_mutations_list:
                        if mutation in noncodings:
                            chosen_indices += [index for index, value in enumerate(sample_mutations_list) if value == mutation]
                
                soonest_mutation = sample_mutations_list[chosen_indices[0]]
                soonest_location = sample_locations_list[chosen_indices[0]]
                chosen_indices.clear()
                sortedcol.append(soonest_mutation)
                location.append(soonest_location)

            return pd.Series([sortedcol, location],index=['mutations', 'locations'])
       
        
        df = combined.apply(sort, axis=1)
        combined['Mutation'] = df['mutations']
        combined['Location'] = df['locations']

        #get a sample_status column that says if the gene has multiple mutations (including dletion and amplification)
        def sample_status(row):
            if len(row['mutations']) > 1: #if there's more than one mutation
                if len(row['mutations']) == 2 and "Wildtype_Tumor" in row['mutations']: #one of the mutations might be a "wildtype tumor"
                    status ="Single_mutation"

                elif len(row['mutations']) == 2 and "Wildtype_Normal" in row['mutations']:
                    status ="Single_mutation"

                else:
                    status = "Multiple_mutation"
            else:
                if row["mutations"] == ["Wildtype_Normal"]:
                    status = "Wildtype_Normal"
                elif row['mutations'] == ['Wildtype_Tumor']:
                    status = "Wildtype_Tumor"
                else:
                    status = "Single_mutation"

            return status
        combined['Mutation_Status'] = combined.apply(sample_status, axis=1)

        #drop all the unnecessary Columns
        df = combined.drop(columns=[mutations_genes+"_CNV", mutations_genes+"_Mutation", mutations_genes+"_Location", mutations_genes+"_Mutation_Status", 'Sample_Status', 'mutations','locations'])
        df['Mutation'] = [','.join(map(str, l)) for l in df['Mutation']]
        df['Location'] = [','.join(map(str, l)) for l in df['Location']]
        if show_location == False: df = df.drop(columns="Location") #if they don't want us to show the location, drop it
        return df


    # Join functions
    def join_omics_to_omics(self, df1_name, df2_name, genes1=None, genes2=None, how="outer", quiet=False, tissue_type="both"):
        """Take specified column(s) from one omics dataframe, and join to specified columns(s) from another omics dataframe. Intersection (inner join) of indices is used.

        Parameters:
        df1_name (str): Name of first omics dataframe to select columns from.
        df2_name (str): Name of second omics dataframe to select columns from.
        genes1 (str, or list or array-like of str, optional): Gene(s) for column(s) to select from df1_name. str if one key, list or array-like of str if multiple. Default of None will select entire dataframe.
        genes2 (str, or list or array-like of str, optional): Gene(s) for Column(s) to select from df2_name. str if one key, list or array-like of str if multiple. Default of None will select entire dataframe.
        how (str, optional): How to perform the join, acceptable values are from ['outer', 'inner', 'left', 'right']. Defaults to 'outer'.
        quiet (bool, optional): Whether to warn when inserting NaNs. Defaults to False.
        tissue_type (str): Acceptable values in ["tumor","normal","both"]. Specifies the desired tissue type desired in the dataframe. Defaults to "both".

        Returns:
        pandas.DataFrame: The selected columns from the two omics dataframes, joined into one dataframe.
        """
        # Check to make sure that the "how" parameter is valid
        self._check_how_parameter(how)

        # Select the columns from each dataframe
        selected1 = self._get_omics_cols(df1_name, genes1, tissue_type)
        selected2 = self._get_omics_cols(df2_name, genes2, tissue_type)

        # Make the multiindices the same
        if selected1.columns.names != selected2.columns.names:
            selected1.columns = add_index_levels(to=selected1.columns, source=selected2.columns)
            selected2.columns = add_index_levels(to=selected2.columns, source=selected1.columns)

        joined = selected1.join(selected2, how= how)

        # Warn them about any NaNs that were inserted in the outer join
        if not quiet and how != "inner":
            self._warn_inserted_nans(df1_name, df2_name, selected1.index, selected2.index)

        # Sort the dataframe so all the tumor samples are first, then all the normal samples
        sample_status_col = self._get_dataframe("clinical")["Sample_Tumor_Normal"]
        joined = sort_df_by_sample_status(joined, sample_status_col)

        return joined

    def join_omics_to_mutations(self, omics_df_name, mutations_genes, omics_genes=None, mutations_filter=None, show_location=True, how="outer", quiet=False, tissue_type="both", mutation_cols = ["Mutation","Location"]):
        """Select all mutations for specified gene(s), and joins them to all or part of the given omics dataframe. Intersection (inner join) of indices is used. Each location or mutation cell contains a list, which contains the one or more location or mutation values corresponding to that sample for that gene, or a value indicating that the sample didn't have a mutation in that gene.

        Parameters:
        omics_df (str): Name of omics dataframe to join the mutation data to.
        mutations_genes (str, or list or array-like of str): The gene(s) to get mutation data for. str if one gene, list or array-like of str if multiple.
        omics_genes (str, or list or array-like of str, optional): Gene(s) to select from the omics dataframe. str if one gene, list or array-like of str if multiple. Default will select entire dataframe.
        mutations_filter (list, optional): List of mutations to prioritize when filtering out multiple mutations, in order of priority. If none of the multiple mutations in a sample are included in mutations_filter, the function will automatically prioritize truncation over missense mutations, and then mutations earlier in the sequence over later mutations. Passing an empty list will cause this default hierarchy to be applied to all samples. Default parameter of None will cause no filtering to be done, and all mutation data will be included, in a list.
        show_location (bool, optional): Whether to include the Location column from the mutation dataframe. Defaults to True.
        how (str, optional): How to perform the join, acceptable values are from ['outer', 'inner', 'left', 'right']. Defaults to 'outer'.
        quiet (bool, optional): Whether to warn when inserting NaNs. Defaults to False.
        tissue_type (str): Acceptable values in ["tumor","normal","both"]. Specifies the desired tissue type desired in the dataframe. Defaults to "both".

        Returns:
        pandas.DataFrame: The mutations for the specified gene, joined to all or part of the omics dataframe. Each location or mutation cell contains a list, which contains the one or more location or mutation values corresponding to that sample for that gene, or a value indicating that the sample didn't have a mutation in that gene.
        """

        # Check to make sure that the "how" parameter is valid
        self._check_how_parameter(how)

        # Select the data from each dataframe
        omics = self._get_omics_cols(omics_df_name, omics_genes, tissue_type)
        mutations = self._get_genes_mutations(mutations_genes, mutations_filter, mutation_cols = mutation_cols)
        if tissue_type == "normal":
            mutations = mutations.iloc[0:0] #If tissue type is normal, we drop all of the mutations rows and join only with the columns.

        mutations_were_filtered = mutations_filter is not None
        joined = self._join_other_to_mutations(omics, mutations, mutations_were_filtered, show_location, how=how, quiet=quiet)

        # Warn them about any NaNs that were inserted in the outer join
        if not quiet and how != "inner":
            self._warn_inserted_nans(omics_df_name, "somatic_mutation", omics.index, mutations.index)

        # Sort the dataframe so all the tumor samples are first, then all the normal samples
        sample_status_col = self._get_dataframe("clinical")["Sample_Tumor_Normal"]
        joined = sort_df_by_sample_status(joined, sample_status_col)

        return joined

    def join_metadata_to_metadata(self, df1_name, df2_name, cols1=None, cols2=None, how="outer", quiet=False, tissue_type="both"):
        """Take specified column(s) from one metadata dataframe, and join to specified columns(s) from another metadata dataframe. Intersection (inner join) of indices is used.

        Parameters:
        df1_name (str): Name of first metadata dataframe to select columns from.
        df2_name (str): Name of second metadata dataframe to select columns from.
        cols1 (str, or list or array-like of str, optional): Column(s) to select from df1_name. str if one key, list or array-like of str if multiple. Default of None will select entire dataframe.
        cols2 (str, or list or array-like of str, optional): Column(s) to select from df2_name. str if one key, list or array-like of str if multiple. Default of None will select entire dataframe.
        how (str, optional): How to perform the join, acceptable values are from ['outer', 'inner', 'left', 'right']. Defaults to 'outer'.
        quiet (bool, optional): Whether to warn when inserting NaNs. Defaults to False.
        tissue_type (str): Acceptable values in ["tumor","normal","both"]. Specifies the desired tissue type desired in the dataframe. Defaults to "both".

        Returns:
        pandas.DataFrame: The selected columns from the two metadata dataframes, joined into one dataframe.
        """
        # Check to make sure that the "how" parameter is valid
        self._check_how_parameter(how)

        # Select the columns from each dataframe
        selected1 = self._get_metadata_cols(df1_name, cols1, tissue_type)
        selected2 = self._get_metadata_cols(df2_name, cols2, tissue_type)

        joined = selected1.join(selected2, how=how, rsuffix='_from_' + df2_name) # Use suffix in case both dataframes have a particular column, such as Patient_ID

        # Warn them about any NaNs that were inserted in the outer join
        if not quiet and how != "inner":
            self._warn_inserted_nans(df1_name, df2_name, selected1.index, selected2.index)

        # Sort the dataframe so all the tumor samples are first, then all the normal samples
        sample_status_col = self._get_dataframe("clinical")["Sample_Tumor_Normal"]
        joined = sort_df_by_sample_status(joined, sample_status_col)

        return joined

    def join_metadata_to_omics(self, metadata_df_name, omics_df_name, metadata_cols=None, omics_genes=None, how="outer", quiet=False, tissue_type="both"):
        """Joins columns from a metadata dataframe (clinical, derived_molecular, or experimental_design) to part or all of an omics dataframe. Intersection (inner join) of indices is used.

        Parameters:
        metadata_df_name (str): Name of metadata dataframe to select columns from.
        omics_df_name (str): Name of omics dataframe to join the metadata columns to.
        metadata_cols (str, or list or array-like of str, optional): Column(s) to select from the metadata dataframe. str if one gene, list or array-like of str if multiple. Default is None, which will select the entire metadata dataframe.
        omics_genes (str, or list or array-like of str, optional): Gene(s) to select data for from the omics dataframe. str if one gene, list or array-like of str if multiple. Default is None, which will select entire dataframe.
        how (str, optional): How to perform the join, acceptable values are from ['outer', 'inner', 'left', 'right']. Defaults to 'outer'.
        quiet (bool, optional): Whether to warn when inserting NaNs. Defaults to False.
        tissue_type (str): Acceptable values in ["tumor","normal","both"]. Specifies the desired tissue type desired in the dataframe. Defaults to "both".

        Returns:
        pandas.DataFrame: The selected metadata columns, joined with all or part of the omics dataframe.
        """
        # Check to make sure that the "how" parameter is valid
        self._check_how_parameter(how)

        # Select the columns from each dataframe
        metadata_selected = self._get_metadata_cols(metadata_df_name, metadata_cols, tissue_type)
        omics_selected = self._get_omics_cols(omics_df_name, omics_genes,tissue_type)

        # Make the indices the same
        if metadata_selected.columns.names != omics_selected.columns.names:
            metadata_selected.columns = add_index_levels(to=metadata_selected.columns, source=omics_selected.columns)

        joined = metadata_selected.join(omics_selected, how=how)

        # Warn them about any NaNs that were inserted in the outer join
        if not quiet and how != "inner":
            self._warn_inserted_nans(metadata_df_name, omics_df_name, metadata_selected.index, omics_selected.index)

        # Sort the dataframe so all the tumor samples are first, then all the normal samples
        sample_status_col = self._get_dataframe("clinical")["Sample_Tumor_Normal"]
        joined = sort_df_by_sample_status(joined, sample_status_col)

        return joined

    def join_metadata_to_mutations(self, metadata_df_name, mutations_genes, metadata_cols=None, mutations_filter=None, show_location=True, how="outer", quiet=False, tissue_type="both"):
        """Select all mutations for specified gene(s), and joins them to all or part of the given metadata dataframe. Intersection (inner join) of indices is used. Each location or mutation cell contains a list, which contains the one or more location or mutation values corresponding to that sample for that gene, or a value indicating that the sample didn't have a mutation in that gene.

        Parameters:
        metadata_df_name (str): Name of metadata dataframe to join the mutation data to.
        mutations_genes (str, or list or array-like of str): The gene(s) to get mutation data for. str if one gene, list or array-like of str if multiple.
        metadata_cols (str, or list or array-like of str, optional): Gene(s) to select from the metadata dataframe. str if one gene, list or array-like of str if multiple. Default will select entire dataframe.
        mutations_filter (list, optional): List of mutations to prioritize when filtering out multiple mutations, in order of priority. If none of the multiple mutations in a sample are included in mutations_filter, the function will automatically prioritize truncation over missense mutations, and then mutations earlier in the sequence over later mutations. Passing an empty list will cause this default hierarchy to be applied to all samples. Default parameter of None will cause no filtering to be done, and all mutation data will be included, in a list.
        show_location (bool, optional): Whether to include the Location column from the mutation dataframe. Defaults to True.
        how (str, optional): How to perform the join, acceptable values are from ['outer', 'inner', 'left', 'right']. Defaults to 'outer'.
        quiet (bool, optional): Whether to warn when inserting NaNs. Defaults to False.
        tissue_type (str): Acceptable values in ["tumor","normal","both"]. Specifies the desired tissue type desired in the dataframe. Defaults to "both".

        Returns:
        pandas.DataFrame: The mutations for the specified gene, joined to all or part of the metadata dataframe. Each location or mutation cell contains a list, which contains the one or more location or mutation values corresponding to that sample for that gene, or a value indicating that the sample didn't have a mutation in that gene.
        """
        # Check to make sure that the "how" parameter is valid
        self._check_how_parameter(how)

        # Select the data from each dataframe
        metadata = self._get_metadata_cols(metadata_df_name, metadata_cols, tissue_type)
        mutations = self._get_genes_mutations(mutations_genes, mutations_filter)
        
        if tissue_type == "normal":
            mutations = mutations.iloc[0:0] #If tissue type is normal, we drop all of the mutations rows and join only with the columns.

        mutations_were_filtered = mutations_filter is not None
        joined = self._join_other_to_mutations(metadata, mutations, mutations_were_filtered, show_location,how=how, quiet=quiet)

        # Warn them about any NaNs that were inserted in the outer join
        if not quiet and how != "inner":
            self._warn_inserted_nans(metadata_df_name, "somatic_mutation", metadata.index, mutations.index)

        # Sort the dataframe so all the tumor samples are first, then all the normal samples
        sample_status_col = self._get_dataframe("clinical")["Sample_Tumor_Normal"]
        joined = sort_df_by_sample_status(joined, sample_status_col)

        return joined
    
    def multi_join(self, join_dict, mutations_filter=None, flatten=False, levels_to_drop=[], how="outer", tissue_type="both"):    
        """Takes a dictionary which keys are dataframes and values are columns from those dataframes and joins all the columns into one dataframe. If the value is an empty list it will join the dataframe
        
        Parameters:
        join_dict (dict): A dictionary with the dataframe and columns to join. Keys are the names of the dataframes and the value is a list of string with the name of the columns corresponding to each dataframe. Example: {'phosphoproteomics':['A2M', 'AAAS'],'proteomics':['AAAS', 'ZZZ3'], 'somatic_mutation':['AHCTF1', 'ZFHX3']}.
        
            Valid dataframes are: acetylproteomics, CNV, phosphoproteomics, phosphoproteomics_gene, proteomics, somatic_mutation_binary, somatic_mutation, transcriptomics, clinical, derived_molecular and experimental_design.
            
            For somatic_mutation_binary it joins all columns that match a gene. Example {'somatic_mutation_binary' : ['A1CF', 'ZYG11B']} It returns a dataframe with all columns that contain those genes.
            
        mutations_filter (list, optional): List of mutations to prioritize when filtering out multiple mutations, in order of priority. If none of the multiple mutations in a sample are included in mutations_filter, the function will automatically prioritize truncation over missense mutations, and then mutations earlier in the sequence over later mutations. Passing an empty list will cause this default hierarchy to be applied to all samples. Default parameter of None will cause no filtering to be done, and all mutation data will be included, in a list.
        
        flatten (bool, optional): Defaults to False and will flatten the multiindexes if set to True.
        
        levels_to_drop (list, optional): Defaults to empty list. Takes a list of strings. Strings are levels to be dropped. If empty it will not drop any.
        
        tissue_type (str): Acceptable values in ["tumor","normal","both"]. Specifies the desired tissue type desired in the dataframe. Defaults to "both"
        
        Returns: pandas.DataFrame
        """
        column_names = []    
        to_join=[]
        
        for df_name in join_dict.keys():
            ## If key belongs to omics
            if df_name in self._valid_omics_dfs:
                # If key is somatic_mutation_binary it will join all columns that match a gene
                if df_name == "somatic_mutation_binary":
                    binary_data = self.get_somatic_mutation_binary()
                    binary_data = "\n".join(list(binary_data.columns))
                    found_genes = []
                    if len(join_dict[df_name]) != 0:
                        for gene in join_dict[df_name]:
                            found_genes += (re.findall((gene+".+\n"), binary_data))
                        found_genes = list(map(lambda x: x.strip("\n"), found_genes))#returns a list of columns that match the given gene
                        columns = self._get_omics_cols(df_name, found_genes, tissue_type= tissue_type)
                    else:
                        columns = self._get_omics_cols(df_name, None, tissue_type= tissue_type)
                else:
                    if len(join_dict[df_name]) != 0:# If there are values to join it will get the columns
                        columns = self._get_omics_cols(df_name, join_dict[df_name], tissue_type = tissue_type)
                    else:# Else join all the dataframe
                        columns = self._get_omics_cols(df_name, None, tissue_type = tissue_type)
            ## If key belongs to metadata 
            elif df_name in self._valid_metadata_dfs:
                if len(join_dict[df_name]) != 0:# If there are values to join it will get the columns
                    columns = self._get_metadata_cols(df_name, join_dict[df_name], tissue_type = tissue_type)
                else:# Else join all the dataframe
                    columns = self._get_metadata_cols(df_name, None, tissue_type = tissue_type)
            ## If key is somatic_mutation 
            elif df_name == "somatic_mutation":
                columns = self._get_genes_mutations(join_dict[df_name], mutations_filter = mutations_filter)

            ### Checks if there are columns with the same name and adds the name of the
            for i in columns.columns:
                if type(i) == tuple:
                    i = reduce(lambda  x, y: str(x)+str(y), i)#returns a flattened column name
                if i in column_names:
                    columns = columns.rename(columns={i: str(i)+'_'+df_name})
                column_names.append(i)
            ###     

            to_join.append(columns)

        joined, how = reduce(self._join_dataframe, to_join, how)
    
        if len(levels_to_drop) != 0:
            joined = ut.reduce_multiindex(joined, levels_to_drop=levels_to_drop)

        if flatten == True:
            joined = ut.reduce_multiindex(joined, flatten=flatten)
        return joined

    # "Private" methods
    
    def _get_dataframe(self, name, tissue_type="both"):
        """Check if a dataframe with the given name exists, and return a copy of it if it does.

        Parameters:
        name (str): The name of the dataframe to get.
        tissue_type (str): Acceptable values in ["tumor","normal","both"]. Specifies the desired tissue type desired in the dataframe. Defaults to "both".

        Returns:
        pandas.DataFrame: A copy of the desired dataframe, if it exists in this dataset.
        """
        if name in self._data.keys():
            df = self._data[name]
            return_df = df.copy(deep=True) # We copy it, with deep=True, so edits on their copy don't affect the master for this instance
            return_df.index.name = df.index.name
            return_df.columns.name = df.columns.name

            if tissue_type == "tumor":
                return self._tumor_only(return_df)
            elif tissue_type == "normal":
                return self._normal_only(return_df)
            elif tissue_type == "both":
                return return_df
            else:
                raise InvalidParameterError(f"Unrecognized value for tissue_type parameter. You passed '{tissue_type}'. Valid options are 'tumor', 'normal', or 'both'.")
        else:
            raise DataFrameNotIncludedError(f"{name} dataframe not included in the {self.get_cancer_type()} dataset.")

    def _get_sample_status_map(self):
        """Get a pandas Series from the clinical dataframe, with sample ids as the index, and each sample's status (tumor or normal) as the values."""
        clinical = self.get_clinical()
        status_map = clinical["Sample_Tumor_Normal"]
        status_map.name = "Sample_Status"
        return status_map

    def _check_df_valid(self, df_name, df_type):
        """Checks whether a dataframe with this name is valid for use as an omics or metadata dataframe in one of the utilties functions. Throws an InvalidParameterError if it isn't.

        Parameters:
        df_name (str): The dataframe name to check.
        df_type (str): Which type of dataframe we're validating--either "omics" or "metadata"

        Returns: None
        """
        if not isinstance(df_name, str): # Check that they passed a str, since utilities functions used to directly accept dataframes
            raise InvalidParameterError(f"Please pass a str for dataframe name parameter. You passed {df_name}, which is a {type(df_name)}")

        if df_type == "omics":
            valid_dfs = self._valid_omics_dfs
        elif df_type == "metadata":
            valid_dfs = self._valid_metadata_dfs
        else:
            raise CptacDevError(f"Invalid df_type of {df_type} passed to cptac.Dataset._check_df_valid.")

        if df_name not in self._data.keys():
            raise DataFrameNotIncludedError(f"{df_name} dataframe not included in the {self.get_cancer_type()} dataset.")
        elif df_name not in valid_dfs:
            error_msg = f"{df_name} is not a valid {df_type} dataframe for this function in this dataset. Valid options:"
            for valid_name in valid_dfs:
                if valid_name in self._data.keys(): # Only print it if it's included in this dataset
                    error_msg = error_msg + '\n\t' + valid_name
            raise InvalidParameterError(error_msg)

    def _warn_inserted_nans(self, name1, name2, index1, index2):
        """Compare two indices from two dataframes, and warn the user that any rows with index values not in both indices were filled with NaNs in a join function.

        Parameters:
        name1 (str): Name of the dataframe the first index came from
        name2 (str): Name of the dataframe the second index came from
        index1 (pandas.Index): First index to compare
        index2 (pandas.Index): Second index to compare

        Returns: None
        """
        unique1 = index1.difference(index2)
        unique2 = index2.difference(index1)

        self._issue_inserted_nans_warning(unique1, name2)
        self._issue_inserted_nans_warning(unique2, name1)

    def _issue_inserted_nans_warning(self, unique, other_name):
        """Issue a warning that the samples in unique were not found in the other_name dataframe, and those column(s) were filled with NaN.

        Parameters:
        unique (list or array-like of str): The samples that weren't in the other_name dataframe.
        other_name (str): The name of the dataframe the samples weren't found in.

        Returns: None
        """
        if other_name == "somatic_mutation":
            return # This will have separate fill warnings printed, because we use different fill values.
        elif len(unique) > 0:
            warnings.warn(f"{other_name} data was not found for the following samples, so {other_name} data columns were filled with NaN for these samples: {', '.join(unique)}", InsertedNanWarning, stacklevel=4)

    def _get_omics_cols(self, omics_df_name, genes, tissue_type="both"):
        """Based on a single gene, or a list or array-like of genes, select multiple columns from an omics dataframe, and return the selected columns as one dataframe.

        Parameters:
        omics_df_name (str): Name of omics dataframe to select column(s) from.
        genes (str, or list or array-like of str): Gene(s) to use to select columns from omics_df. str if one gene, list or array-like if multiple. Passing None will select the entire omics dataframe.
        tissue_type (str): Acceptable values in ["tumor","normal","both"]. Specifies the desired tissue type desired in the dataframe. Defaults to "both".

        Returns:
        pandas.DataFrame: The selected columns from the dataframe.
        """
        # Check that they passed a valid omics df
        self._check_df_valid(omics_df_name, "omics")

        # Get our omics df, using _get_dataframe to catch invalid requests
        omics_df = self._get_dataframe(omics_df_name, tissue_type)

        # Process genes parameter
        if isinstance(genes, str): # If it's a single gene, make it a list so we can treat everything the same
            genes = [genes]
        elif isinstance(genes, (list, pd.Series, pd.Index)): # If it's already a list or array-like, we're all good
            pass
        elif genes is None: # If it's the default of None, rename columns and return the entire dataframe
            # Add the gene name to end beginning of each column header, to preserve info when we join dataframes.
            if isinstance(omics_df.columns, pd.MultiIndex):
                omics_df.columns = omics_df.columns.set_levels(omics_df.columns.levels[0] + '_' + omics_df_name, level=0)
            else:
                omics_df = omics_df.add_suffix('_' + omics_df_name)
            return omics_df
        else: # If it's none of those, they done messed up. Tell 'em.
            raise InvalidParameterError("Genes parameter \n{}\nis of invalid type {}. Valid types: str, list or array-like of str, or NoneType.".format(genes, type(genes)))

        genes = pd.Index(genes, name="Name")

        if isinstance(omics_df.columns, pd.MultiIndex):
            contained = genes.intersection(omics_df.columns.get_level_values("Name")).drop_duplicates() # Get the genes that actually exist in the dataframe's columns
            mi_contained = omics_df.columns[omics_df.columns.get_level_values("Name").isin(genes)]

            not_contained = genes.difference(contained).drop_duplicates() # So we can warn the user later
            arrays = [not_contained] + [[np.nan] for i in range(omics_df.columns.nlevels - 1)]
            mi_not_contained = pd.MultiIndex.from_product(arrays, names=omics_df.columns.names)

            genes = mi_contained.union(mi_not_contained) # To use for reindexing the dataframe
        else:
            contained = genes.intersection(omics_df.columns).drop_duplicates() # Get the genes that actually exist in the dataframe's columns
            not_contained = genes.difference(contained).drop_duplicates() # So we can warn the user later

        selected = omics_df[contained]
        selected = selected.reindex(columns=genes) # This will add the columns not included in the dataframe, and fill them with NaN.

        # Warn the user about columns filled with NaN
        if len(not_contained) > 0:
            warnings.warn(f"The following columns were not found in the {omics_df_name} dataframe, so they were inserted into joined table, but filled with NaN: {', '.join(not_contained)}", ParameterWarning, stacklevel=3)

        # Append dataframe name to end of each column header, to preserve info when we merge dataframes
        if isinstance(omics_df.columns, pd.MultiIndex):
            selected.columns = selected.columns.set_levels(selected.columns.levels[0] + '_' + omics_df_name, level=0)
        else:
            selected = selected.add_suffix('_' + omics_df_name)

        selected.columns.name = "Name"
        return selected

    def _get_metadata_cols(self, df_name, cols, tissue_type="both"):
        """Select a single column or several columns from a metadata dataframe.

        Parameters:
        df_name (str): The name of the metadata dataframe to select the column(s) from.
        cols (str, or list or array-like of str): The column(s) to select from the dataframe. str if single, list or array-like of str if multiple. Passing None will select the entire dataframe.
        tissue_type (str): Acceptable values in ["tumor","normal","both"]. Specifies the desired tissue type desired in the dataframe. Defaults to "both".

        Returns:
        pandas.DataFrame: The specified columns from the given dataframe.
        """
        # Check that they passed a valid metadata df
        self._check_df_valid(df_name, "metadata")

        # Get our dataframe, using _get_dataframe to catch invalid requests
        df = self._get_dataframe(df_name, tissue_type)

        # Process genes parameter
        if isinstance(cols, str): # If it's a single column, make it a list so we can treat everything the same
            cols = [cols]
        elif isinstance(cols, (list, pd.Series, pd.Index)): # If it's already a list or array-like, we're all good
            pass
        elif cols is None: # If it's the default of None, return the entire dataframe
            return df
        else: # If it's none of those, they done messed up. Tell 'em.
            raise InvalidParameterError("Columns parameter {} is of invalid type {}. Valid types: str, or list or array-like of str.".format(cols, type(cols)))

        cols = pd.Index(cols).drop_duplicates()

        # Check that they didn't pass any invalid columns
        not_contained = cols.difference(df.columns)
        if len(not_contained) > 0:
            raise InvalidParameterError(f'The following columns were not found in the {df_name} dataframe: {", ".join(not_contained)}')

        selected = df[cols]
        return selected

    def _get_genes_mutations(self, genes, mutations_filter ,mutation_cols = ["Mutation","Location"]):
            """Gets all the mutations for one or multiple genes, for all patients.

            Parameters:
            genes (str, or list or array-like of str): The gene(s) to grab mutations for. str if one, list or array-like of str if multiple.
            mutations_filter (list, optional): List of mutations to prioritize when filtering out multiple mutations, in order of priority. If none of the multiple mutations in a sample are included in mutations_filter, the function will automatically prioritize truncation over missense mutations, and then mutations earlier in the sequence over later mutations. Passing an empty list will cause this default hierarchy to be applied to all samples. Passing None will cause no filtering to be done, and all mutation data will be included, in a list.
            mutation_cols (list or str): List of columns to include in joined df. Default is the Mutation and Location column. The str "All" returns all mutation columns. 

            Returns:
            pandas.DataFrame: The mutations in each patient for the specified gene(s).
            """
             
            somatic_mutation = self.get_somatic_mutation()                            

            # Process genes parameter
            if isinstance(genes, str): # If it's a single gene, make it a list so we can treat everything the same
                genes = [genes]
            elif isinstance(genes, (list, pd.Series, pd.Index)): # If it's already a list or array-like, we're all good
                pass
            else: # If it's neither of those, they done messed up. Tell 'em.
                raise InvalidParameterError("Genes parameter {} is of invalid type {}. Valid types: str, or list or array-like of str.".format(genes, type(genes)))

            # Set some column names for use later
            gene_col = "Gene"
            mutation_col = "Mutation"
            location_col = "Location"
            mutation_status_col = "Mutation_Status"

            # Check that they didn't make any typos in specifying filter values
            invalid_filter = False
            if mutations_filter is not None:
                for filter_val in mutations_filter:
                    if (filter_val not in somatic_mutation[mutation_col].values) and (filter_val not in somatic_mutation[location_col].values):
                        raise InvalidParameterError(f"Filter value {filter_val} does not exist in the mutations dataframe for this dataset. Check for typos and existence. Merge aborted.")

            # Create an empty dataframe, which we'll fill with the columns we select using our genes, and then return.
            df = pd.DataFrame(index=somatic_mutation.index.copy().drop_duplicates())
            genes = pd.Series(genes).drop_duplicates()
            for gene in genes:
                gene_mutations = somatic_mutation[somatic_mutation[gene_col] == gene] # Get all the mutations for that gene
                if len(gene_mutations) == 0: # If the gene doesn't match any genes in the dataframe, tell them
                    raise InvalidParameterError("{} gene not found in somatic_mutation data.".format(gene))
                gene_mutations = gene_mutations.drop(columns=[gene_col]) # Gene column is same for every sample, so we don't need it anymore.

                # Check whether all filter values exist for this particular gene. If not, that's fine, we just want to warn the user.
                if mutations_filter is not None:
                    for filter_val in mutations_filter:
                        if (filter_val not in gene_mutations[mutation_col].values) and (filter_val not in gene_mutations[location_col].values):
                            warnings.warn(f"Filter value {filter_val} does not exist in the mutations data for the {gene} gene, though it exists for other genes.", ParameterWarning, stacklevel=3)

                # Create another empty dataframe, which we'll fill with the mutation and location data for this gene, as lists
                prep_index = gene_mutations.index.copy().drop_duplicates()
                prep_columns = gene_mutations.columns.copy()
                mutation_status_idx = pd.Index([mutation_status_col]) # Prep mutation_status_col to be joined
                prep_cols_with_mut_status = prep_columns.append(mutation_status_idx) # Add a mutation_status column, which will indicate if there are 1 or multiple mutations
                mutation_lists = pd.DataFrame(index=prep_index, columns=prep_cols_with_mut_status)
                
                if mutation_cols == "All":
                    mutation_cols = somatic_mutation.columns.to_list()
                    mutation_cols = mutation_cols[1:] #drop column name "Gene"  
                                        
                # Get the mutation(s), mutation status, and location information for this gene and sample
                # Yes, I know I'm doing that horrible thing, using nested for loops to work with dataframes. However, I tried refactoring it to use DataFrame.groupby and DataFrame.apply, and both actually made it slower. Go figure.
                for sample in mutation_lists.index: # samples ids with mutation

                    sample_data = gene_mutations.loc[sample] # Get slice of dataframe for the sample

                    for col in mutation_cols:

                        sliced_col = sample_data[col] # Get single col 

                        #Make Mutations List 
                        if col == mutation_col:
                            if isinstance(sliced_col, pd.Series):
                                sample_mutations_list = sliced_col.tolist()
                            else:
                                sample_mutations_list = [sliced_col] 

                            # Figure out what our mutation status is (either single_mutation or multiple_mutation)
                            if len(sample_mutations_list) > 1:
                                sample_mutation_status = "Multiple_mutation"
                            else:
                                sample_mutation_status = "Single_mutation"
                        #Make Location List        
                        if col == location_col:
                            if isinstance(sliced_col, pd.Series):
                                sample_locations_list = sliced_col.tolist()
                            else:
                                sample_locations_list = [sliced_col]
                        else:
                            if isinstance(sliced_col, pd.Series):
                                sample_mut_col_list = sliced_col.tolist()
                            else:
                                sample_mut_col_list = [sliced_col]

                            mutation_lists.at[sample, col] = sample_mut_col_list  
                                        
                    necessary_cols = ["Mutation","Location"] #must be included in order to filter mutations                   
                    if mutations_filter is not None: # Filter multiple mutations down to just one
                        #Check that Mutation and Location is mutation_cols so mutations_filter can work 
                        if (all(x in mutation_cols for x in necessary_cols)):
                            chosen_mutation, chosen_location = self._filter_multiple_mutations(mutations_filter, sample_mutations_list, sample_locations_list)
                            mutation_lists.at[sample, mutation_col] = chosen_mutation
                            mutation_lists.at[sample, location_col] = chosen_location
                        
                    else: # Include all the mutations!
                        if (all(x in mutation_cols for x in necessary_cols)):
                            mutation_lists.at[sample, mutation_col] = sample_mutations_list
                            mutation_lists.at[sample, location_col] = sample_locations_list 


                    # Also add the mutations status column
                    mutation_lists.at[sample, mutation_status_col] = sample_mutation_status
                                        
                # Print warning if either Location and Mutation not provided and mutation_filter provided                       
                if (mutations_filter is not None) & ((all(x in mutation_cols for x in necessary_cols)) == False):
                    warnings.warn(f"mutations_filter was not applied because columns 'Location' and 'Mutation' were not included in mutation_cols.", ParameterWarning, stacklevel=3)                        

                mutation_lists = mutation_lists[mutation_cols + [mutation_status_col]] #only include user specified columns and Mutation_Status 
                mutation_lists = mutation_lists.add_prefix(gene + '_') # Add the gene name to end beginning of each column header, to preserve info when we join dataframes.
                df = df.join(mutation_lists, how='outer') # Append the columns to our dataframe we'll return.
                df.columns.name = "Name"

            return df
                                        
    def _join_other_to_mutations(self, other, mutations, mutations_were_filtered, show_location, how, quiet):
        """Join selected mutations data to selected other omics or metadata, add a Sample_Status column, fill in NaNs with Wildtype_Normal or Wildtype_Tumor, and name the dataframe.

        Parameters:
        other (pandas.DataFrame): The selected data from the other type of dataframe (omics or metadata) to join with the selected mutations.
        mutations (pandas.DataFrame): The selected mutations data to join with.
        mutations_were_filtered (bool): Whether multiple mutations in the mutations data were filtered down to just one, or not. Determines whether fill values are in lists or not.
        show_location (bool): Whether to include the Location column from the mutation dataframe.
        quiet (bool): Whether to show warning when filling in rows with no mutation data with "Wildtype_Tumor" or "Wildtype_Normal"

        Returns:
        pandas.DataFrame: The joined dataframe, with a Sample_Status column added and NaNs filled.
        """
        # Make the indices the same
        if mutations.columns.nlevels != other.columns.nlevels:
            mutations.columns = add_index_levels(to=mutations.columns, source=other.columns)
        joined = other.join(mutations, how=how)

        # Add Sample_Status column by joining the sample_status_map to the joined mutation dataframe.
        sample_status_map = self._get_sample_status_map()
        joined = join_col_to_dataframe(joined, sample_status_map)
        joined.columns.name = "Name" # This attribute gets lost in the join above

        # Set our fill values
        wildtype_normal_fill = "Wildtype_Normal"
        wildtype_tumor_fill = "Wildtype_Tumor"
        no_mutation_fill = "No_mutation"

        # Fill in Wildtype_Normal or Wildtype_Tumor for NaN values (i.e., no mutation data for that sample) in joined dataframe mutation columns
        mutation_regex = r'^.*_Mutation$' # Construct regex to find all mutation columns
        mutation_cols = joined.columns[joined.columns.get_level_values("Name").str.match(mutation_regex)] # Get a list of all mutation columns

        fill_log = [] # We're going to keep track of value filling, and let the user know we did it.
        for mutation_col in mutation_cols:

            # Log how many values we're going to fill for this gene
            num_filled = (((sample_status_map == "Normal") | (sample_status_map == "Tumor")) & (pd.isnull(joined[mutation_col]))).sum() # See how many values we'll fill by using sum to get number of "True" in array
            if num_filled > 0:
                if isinstance(mutation_col, tuple):
                    gene = mutation_col[0].rsplit("_", maxsplit=1)[0]
                else:
                    gene = mutation_col.rsplit("_", maxsplit=1)[0]
                fill_log.append(f"{num_filled} samples for the {gene} gene")

            # Impute values
            joined.loc[(sample_status_map == "Normal") & (pd.isnull(joined[mutation_col])), mutation_col] = wildtype_normal_fill # Change all NaN mutation values for Normal samples to Wildtype_Normal.
            joined.loc[(sample_status_map == "Tumor") & (pd.isnull(joined[mutation_col])), mutation_col] = wildtype_tumor_fill # Change all NaN mutation values for Tumor samples to Wildtype_Tumor

            # If we didn't filter mutations, encapsulate the fill values in lists, to match the other values in the column
            if not mutations_were_filtered:
                joined[mutation_col] = joined[mutation_col].apply(lambda x: x if isinstance(x, list) else [x])

        if len(fill_log) > 0 and not quiet:
            warnings.warn(f"In joining the somatic_mutation table, no mutations were found for the following samples, so they were filled with Wildtype_Tumor or Wildtype_Normal: {', '.join(fill_log)}", FilledMutationDataWarning, stacklevel=3)

        # Depending on show_location, either fill NaN values in the joined dataframe location columns with "No_mutation", or just drop the location columns altogether
        location_regex = r'^.*_Location$' # Construct regex to find all location columns
        location_cols = joined.columns[joined.columns.get_level_values("Name").str.match(location_regex)] # Get a list of all location columns
        for location_col in location_cols:
            if show_location: # If we're including the location column, fill NaN with "No_mutation", since that's what it means, so things are clearer to the user.

                # Make sure Sample Status is not NaN, though--if it is, we have no mutation data at all for that sample, so we can't say "No_mutation". It must have been a sample that was in the other dataframe, but not the mutations.
                joined.loc[(pd.isnull(joined[location_col])) & (pd.notnull(sample_status_map)), location_col] = no_mutation_fill 
                
                # If we didn't filter mutations, encapsulate the fill values in lists, to match the other values in the column
                if not mutations_were_filtered:
                    joined[location_col] = joined[location_col].apply(lambda x: x if isinstance(x, list) else [x])
            else:
                joined = joined.drop(columns=[location_col]) # Drop the location column, if the caller wanted us to.

        # Fill NaN values in Mutation_Status column with either Wildtype_Tumor or Wildtype_Normal
        mutation_status_regex = r"^.*_Mutation_Status$" # Construct a regex to find all Mutation_Status columns
        mutation_status_cols = joined.columns[joined.columns.get_level_values("Name").str.match(mutation_status_regex)] # Get a list of all Mutation_Status columns
        for mutation_status_col in mutation_status_cols:
            joined.loc[(sample_status_map == "Normal") & (pd.isnull(joined[mutation_status_col])), mutation_status_col] = "Wildtype_Normal" # Change all NaN mutation status values for Normal samples to Wildtype_Normal
            joined.loc[(sample_status_map == "Tumor") & (pd.isnull(joined[mutation_status_col])), mutation_status_col] = "Wildtype_Tumor" # Change all NaN mutation status values for Tumor samples to Wildtype_Tumor

        return joined

    def _filter_multiple_mutations(self, mutations_filter, sample_mutations_list, sample_locations_list):
        """Based on a mutations filter, choose one mutation and its location from two lists of mutations and locations.

        Parameters:
        mutations_filter (list of str): A list of mutations to prioritize, in order of priority. Passing an empty list will cause truncations to be chosen over missense, and mutations earlier in the sequence over later ones.
        sample_mutations_list (list of str): The mutations to filter.
        sample_locations_list (list of str): The locations to filter, in the same order as the mutations.

        Returns:
        str: The chosen mutation
        str: The chosen location
        """
        # Based on the cancer type, define which mutation types are truncations, for sorting later
        if self._cancer_type == 'colon':
            truncations = ['frameshift deletion', 'frameshift insertion', 'frameshift substitution', 'stopgain', 'stoploss']
            missenses = ['nonframeshift deletion', 'nonframeshift insertion', 'nonframeshift substitution', 'nonsynonymous SNV']
        elif self._cancer_type == 'hnscc' and self.version() == "0.1":
            truncations =["stopgain", "stoploss"]
            missenses = ["nonframeshift insertion", "nonframeshift deletion"]
        else:
            truncations = ['Frame_Shift_Del', 'Frame_Shift_Ins', 'Nonsense_Mutation', 'Nonstop_Mutation', 'Splice_Site']
            missenses = ['In_Frame_Del', 'In_Frame_Ins', 'Missense_Mutation']

        if self._cancer_type == "gbm":
            noncodings = ["Intron", "RNA", "3'Flank", "Splice_Region", "5'UTR", "5'Flank", "3'UTR"]

        # Filter the mutations!!
        chosen_indices = []
        for filter_val in mutations_filter: # This will start at the beginning of the filter list, thus filters earlier in the list are prioritized, like we want
            if filter_val in sample_mutations_list:
                chosen_indices = [index for index, value in enumerate(sample_mutations_list) if value == filter_val]
            elif filter_val in sample_locations_list: # The mutations filter can also filter by location
                chosen_indices = [index for index, value in enumerate(sample_locations_list) if value == filter_val]
            if len(chosen_indices) > 0: # We found at least one mutation from the filter to prioritize, so we don't need to worry about later values in the filter priority list
                break

        if len(chosen_indices) == 0: # None of the mutations for the sample were in the filter, so we're going to have to use our default hierarchy
            for mutation in sample_mutations_list:
                if mutation in truncations:
                    chosen_indices += [index for index, value in enumerate(sample_mutations_list) if value == mutation]

        if len(chosen_indices) == 0: # None of them were in the filter, nor were truncations, so we'll grab all the missenses
            for mutation in sample_mutations_list:
                if mutation in missenses:
                    chosen_indices += [index for index, value in enumerate(sample_mutations_list) if value == mutation]

        if self._cancer_type == "gbm" and len(chosen_indices) == 0: # None of them were in the filter, nor were truncations, nor missenses, so we'll grab all the noncodings
            for mutation in sample_mutations_list:
                if mutation in noncodings:
                    chosen_indices += [index for index, value in enumerate(sample_mutations_list) if value == mutation]

        if len(chosen_indices) == 0: # There were no truncations or missenses, so they should all be Silent mutations
            for mutation in sample_mutations_list:
                if mutation not in ["Silent", "synonymous SNV"]:
                    warnings.warn(f"Unknown mutation type {mutation}. Assigned lowest priority in filtering.", ParameterWarning, stacklevel=4)
            chosen_indices = range(len(sample_mutations_list)) # We'll sort them all by location

        # If there are multiple mutations in chosen_indices, the following code will pick the one soonest in the peptide sequence.
        soonest_mutation = sample_mutations_list[chosen_indices[0]]
        soonest_location = sample_locations_list[chosen_indices[0]]
        for index in chosen_indices:
            mutation = sample_mutations_list[index]
            location = sample_locations_list[index]

            if pd.isnull(location): # Some of the mutations have no location. We'll de-prioritize those.
                continue
            if pd.isnull(soonest_location): # This would happen if our initial value for soonest_location was NaN. If we got here, then the one we're testing isn't null, and we'll automatically prefer it
                soonest_location = location
                soonest_mutation = mutation
                continue

            num_location = self._parse_mutation_location(location) # Here, we're parsing the numerical position out of the location strings, for comparisons
            num_soonest_location = self._parse_mutation_location(soonest_location)
            if num_location < num_soonest_location:
                soonest_location = location
                soonest_mutation = mutation
        return soonest_mutation, soonest_location

    def _parse_mutation_location(self, location):
        """Parse the number out of the location for a mutation.

        Parameters:
        location (str): The location to parse.

        Returns:
        int: The numerical part of the location.
        """
        if pd.isnull(location):
            return location
        num = ""
        found_digits = False
        for char in location:
            if char.isdigit():
                num = num + char
                found_digits = True
            else:
                if found_digits: # We only want the first block of numbers
                    return int(num)
        return int(num) # We get here if the location ended with a digit

    def _tumor_only(self, df):
        """For a given dataframe, extract only the tumor samples."""
        
        tumor_df = df[~df.index.str.endswith(".N")]                  
        return tumor_df

    def _normal_only(self, df):
        """For a given dataframe, extract only the tumor samples."""

        normal_df = df[df.index.str.endswith(".N")]
        return normal_df

    def _check_how_parameter(self, given_how):
        possible_values = ['outer', 'inner', 'left', 'right']
        if given_how not in possible_values:
            raise InvalidParameterError("'{}' is not a valid value for 'how'. Possible values are 'outer', 'inner', 'left', 'right'.".format(given_how))
            
    def _join_dataframe(self, df1, df2):
        """Joins a dataframe to another dataframe.
 
        Returns: pandas.DataFrame
        """
        if type(df1) == str:
            return [df2, df1]
        elif type(df1) == list:
            how = df1[1]
            if df1[0].columns.names != df2.columns.names:
                    df1[0].columns = add_index_levels(to=df1[0].columns, source=df2.columns)
                    df2.columns = add_index_levels(to=df2.columns, source=df1[0].columns)

            joined = df1[0].join(df2, how= how)
        return [joined,how]



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import os
import warnings
import datetime

from cptac.dataset import Dataset
from cptac.dataframe_tools import *
from cptac.exceptions import FailedReindexWarning, PublicationEmbargoWarning, ReindexMapError

class PdcLuad(Dataset):

    def __init__(self, version="latest", no_internet=False):
        """Load all of the dataframes as values in the self._data dict variable, with names as keys, and format them properly.

        Parameters:
        version (str, optional): The version number to load, or the string "latest" to just load the latest building. Default is "latest".
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """

        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        # This keeps a record of all versions that the code is equipped to handle. That way, if there's a new data release but they didn't update their package, it won't try to parse the new data version it isn't equipped to handle.
        valid_versions = ["1.0"]

        data_files = {
            "1.0": [
                "acetylome.tsv.gz",
                "clinical.tsv.gz",
                "phosphoproteome.tsv.gz",
                "proteome.tsv.gz",
                "aliquot_to_patient_ID.tsv"
            ]
        }

        # Call the parent class __init__ function
        super().__init__(cancer_type="pdcluad", version=version, valid_versions=valid_versions, data_files=data_files, no_internet=no_internet, attempt_update_index=False)

        # Load the data into dataframes in the self._data dict
        loading_msg = f"Loading {self.get_cancer_type()} v{self.version()}"
        for file_path in self._data_files_paths: # Loops through files variable

            # Print a loading message. We add a dot every time, so the user knows it's not frozen.
            loading_msg = loading_msg + "."
            print(loading_msg, end='\r')

            path_elements = file_path.split(os.sep) # Get a list of the levels of the path
            file_name = path_elements[-1] # The last element will be the name of the file. We'll use this to identify files for parsing in the if/elif statements below

            if file_name == "clinical.tsv.gz":
                df = pd.read_csv(file_path, sep="\t", index_col=0)
                clin_drop_rows = ['Internal Reference - Pooled Sample', 'Normal Only IR', 'Taiwanese IR', 'Tumor Only IR']
                df = df.drop(clin_drop_rows, axis = 'index') # drop quality control and ref intensity
                self._data["clinical"] = df

            if file_name == "acetylome.tsv.gz":
                df = pd.read_csv(file_path, sep="\t")
                self._data["acetylproteomics"] = df

            if file_name == "phosphoproteome.tsv.gz":
                df = pd.read_csv(file_path, sep="\t")
                self._data["phosphoproteomics"] = df

            if file_name == "proteome.tsv.gz":
                df = pd.read_csv(file_path, sep="\t")
                self._data["proteomics"] = df
                
            # aliquot_to_patient_ID.tsv contains only unique aliquots (no duplicates), 
            # so there is no need to slice out cancer specific aliquots
            elif file_name == "aliquot_to_patient_ID.tsv":
                df = pd.read_csv(file_path, sep = "\t", index_col = 'aliquot_ID', usecols = ['aliquot_ID', 'patient_ID'])
                map_dict = df.to_dict()['patient_ID'] # create dictionary with aliquot_ID as keys and patient_ID as values
                self._helper_tables["map_ids"] = map_dict

        print(' ' * len(loading_msg), end='\r') # Erase the loading message
        formatting_msg = f"Formatting {self.get_cancer_type()} dataframes..."
        print(formatting_msg, end='\r')
        
        # There were 2 duplicate IDs (same case ID and aliquot) in the phosphoproteomics and acetylproteomics data.
        # I used the mapping file "aliquot_to_patient_ID.tsv" to determine the tissue type for these duplicates. 
        # They were all tumor samples. Next, I ran a pearson correlation to check how well the values from each 
        # aliquot correlated to its respective flagship sample.  I also created a scatterplot for each aliquot 
        # and flagship pair. For phosphoproteomics, each aliquot correlated well with its flagship sample
        # so we averaged them. For acetylproteomics, the second occurrence of the duplicates correlated better 
        # with the flagship values, so we dropped the first ocurrence. 
        # A file containing the correlations can be downloaded at: 
        # https://byu.box.com/shared/static/jzsq69bd079oq0zbicw4w616hyicd5ev.xlsx
        
        # Common rows to drop
        drop_rows = ['Normal Only IR', 'Taiwanese IR', 'Tumor Only IR']

        # Get dictionary with aliquot_ID as keys and patient_ID as values
        mapping_dict = self._helper_tables["map_ids"]
        
        # Proteomics
        prot = self._data["proteomics"]
        prot['Patient_ID'] = prot['aliquot_submitter_id'].replace(mapping_dict) # map aliquots to patient IDs (normals have '.N')
        prot = prot.set_index('Patient_ID')
        prot = prot.drop(['aliquot_submitter_id', 'case_submitter_id'], axis = 'columns')
        self._data["proteomics"] = prot
        
        # Phosphoproteomics
        phos = self._data["phosphoproteomics"]
        phos['Patient_ID'] = phos['aliquot_submitter_id'].replace(mapping_dict) # map aliquots to patient IDs (normals have '.N')
        phos = phos.set_index('Patient_ID')
        phos = phos.drop(['aliquot_submitter_id', 'case_submitter_id'], axis = 'columns') 
        phos = phos.drop(drop_rows, axis = 'index')
        # these duplicates had high correlations with their tumor flagship samples, so we average them (see long comment above)
        phos = average_replicates(phos, ['C3N-02379', 'C3N-02587']) 
        phos = map_database_to_gene_pdc(phos, 'refseq') # map refseq IDs to gene names
        self._data["phosphoproteomics"] = phos
        
        # Acetylproteomics
        acetyl = self._data["acetylproteomics"]
        acetyl['Patient_ID'] = acetyl['aliquot_submitter_id'].replace(mapping_dict) # aliquots to patient IDs (normals have '.N')
        acetyl = acetyl.set_index('Patient_ID')
        acetyl = acetyl.drop(['aliquot_submitter_id', 'case_submitter_id'], axis = 'columns') 
        acetyl = acetyl.drop(drop_rows, axis = 'index')
        acetyl = map_database_to_gene_pdc(acetyl, 'refseq') # map refseq IDs to gene names
        acetyl = rename_duplicate_labels(acetyl, 'index') # add ".1" to the second ocurrence of the IDs with duplicates
        # drop 1st occurrence - didn't correlate as well with flagship
        acetyl = acetyl.drop(['C3N-02379', 'C3N-02587'], axis = 'index') 
        acetyl = acetyl.rename(index = {'C3N-02379.1':'C3N-02379', 'C3N-02587.1': 'C3N-02587'})# delete '.1' duplicate identifier 
        self._data["acetylproteomics"] = acetyl

        
        # Sort rows (tumor first then normal) and columns by first level (protein/gene name)
        self._data = sort_all_rows_pancan(self._data)  
        

        print(" " * len(formatting_msg), end='\r') # Erase the formatting message



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import os
import warnings
import datetime

from cptac.dataset import Dataset
from cptac.dataframe_tools import *
from cptac.exceptions import FailedReindexWarning, PublicationEmbargoWarning, ReindexMapError

class PdcBrca(Dataset):

    def __init__(self, version="latest", no_internet=False):
        """Load all of the dataframes as values in the self._data dict variable, with names as keys, and format them properly.

        Parameters:
        version (str, optional): The version number to load, or the string "latest" to just load the latest building. Default is "latest".
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """

        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        # This keeps a record of all versions that the code is equipped to handle. That way, if there's a new data release but they didn't update their package, it won't try to parse the new data version it isn't equipped to handle.
        valid_versions = ["1.0"]

        data_files = {
            "1.0": [
                "acetylome.tsv.gz",
                "clinical.tsv.gz",
                "phosphoproteome.tsv.gz",
                "proteome.tsv.gz",
            ]
        }

        # Call the parent class __init__ function
        super().__init__(cancer_type="pdcbrca", version=version, valid_versions=valid_versions, data_files=data_files, no_internet=no_internet, attempt_update_index=False)

        # Load the data into dataframes in the self._data dict
        loading_msg = f"Loading {self.get_cancer_type()} v{self.version()}"
        for file_path in self._data_files_paths: # Loops through files variable

            # Print a loading message. We add a dot every time, so the user knows it's not frozen.
            loading_msg = loading_msg + "."
            print(loading_msg, end='\r')

            path_elements = file_path.split(os.sep) # Get a list of the levels of the path
            file_name = path_elements[-1] # The last element will be the name of the file. We'll use this to identify files for parsing in the if/elif statements below

            if file_name == "clinical.tsv.gz":
                df = pd.read_csv(file_path, sep="\t", index_col=0)
                df = df.drop(['Internal Reference - Pooled Sample', 'RetroIR'])
                df = df.rename(index={'604':'CPT000814'}) # use the aliquot for 604
                self._data["clinical"] = df

            if file_name == "acetylome.tsv.gz":
                df = pd.read_csv(file_path, sep="\t")
                df = df.set_index(["case_submitter_id", "aliquot_submitter_id"]) 
                df = df.drop('RetroIR', level = 1)
                self._data["acetylproteomics"] = df

            if file_name == "phosphoproteome.tsv.gz":
                df = pd.read_csv(file_path, sep="\t")
                df = df.set_index(["case_submitter_id", "aliquot_submitter_id"])
                df = df.drop('RetroIR', level = 1)
                self._data["phosphoproteomics"] = df

            if file_name == "proteome.tsv.gz":
                df = pd.read_csv(file_path, sep="\t")
                df = df.set_index(["case_submitter_id", "aliquot_submitter_id"])
                self._data["proteomics"] = df

                
        print(' ' * len(loading_msg), end='\r') # Erase the loading message
        formatting_msg = f"Formatting {self.get_cancer_type()} dataframes..."
        print(formatting_msg, end='\r')
        
        # prosp-brca-all-samples.txt shows which patient IDs have normal samples and which have replicates. This file
        # can be found on Box under CPTAC/cptac/pancan/helper_files 
        # 7 IDs with replicates: '11BR031', '11BR053', '11BR036', '11BR060', '14BR005', '11BR011', '21BR010'
        replicates = ['11BR031', '11BR053', '11BR036', '11BR060', '14BR005', '11BR011', '21BR010']
        
        # Drop normal aliquots 
        # There were 2 aliquots for 17 patients in the proteomics data. I calculated pearson correlations 
        # and created scatterplots to compare the values from each aliquot to its respective patient_ID in the 
        # original cptac data (flagship values). One aliquot for each patient did not correlate well 
        # (correlations between 0.001 and 0.4), while the other did (correlations between 0.7 and 0.9). 
        # Karsten Krug from the Broad suggested that the aliquots that did not correlate well are likely normal samples 
        # (18 patients had normal samples) which were dropped in downstream analysis because of quality control issues. 
        # The Patient_IDs with 2 aliquots did have a normal sample run (see mapping file 'prosp-brca-all-samples.txt').
        # Therefore, we drop them here. We kept the other aliquots that did correlate well. Of the 18 IDs with normal 
        # samples, 21BR010 only had one aliquot and it did not correlate well so it was dropped. I checked that the 
        # aliquots we dropped were normal samples using the biospecimen manifest for Brca proteomics on the PDC website 
        # under the biospecimen tab. I checked that the normal aliquots are the same for all omics, so we also dropped
        # them for phosphoproteomics and acetylproteomics. 
        # A file containing the correlations can be downloaded at: 
        # https://byu.box.com/shared/static/jzsq69bd079oq0zbicw4w616hyicd5ev.xlsx
        
        drop_normals = ['64ee175f-f3ce-446e-bbf4-9b6fa8_D1', '7ac27de9-0932-4ff5-aab8-29c527',
            '3208e021-1dae-42fd-bd36-0f3c3d', '6c660b6b-bfda-47b0-9499-160d49','241ecd0e-89bd-4d3a-81b3-55a250',
            '428de0d4-7f84-4075-bae1-352af6', '0a80d3c4-0758-447a-958c-ea868c', '53723086-8858-4395-93d7-0baa68',
            '1740224c-32d1-4c9f-98c6-653363', '885fe794-a98e-4f81-a284-ac4bb8', '4749ba99-d3b8-4ae3-b6f6-458bc7',
            '81116212-b7e6-454b-9579-105cf3', '1664b920-5e60-4e3b-9aab-fe121c', '3367406e-d39c-4641-a3e7-44e1f3',
            'e3d45dc6-66ef-4e0b-9d96-1b5db5', '33adae13-5dbd-4530-a5d5-3763e4', 'acf022b3-7f01-43b3-ac14-86f97d',
            '39f81c85-1832-45eb-829a-3040ad']
        
        # Proteomics
        prot = self._data["proteomics"]
        prot = prot.drop(drop_normals, level = 'aliquot_submitter_id') # drop normal aliquots (QC issues)
        prot = prot.rename(index={'604':'CPT000814'}) # use the aliquot for 604 
        prot.index = prot.index.droplevel('aliquot_submitter_id')
        self._data["proteomics"] = prot
        
        # Note -> At the time this code was added, phospho had all NaN values. In an email recieved 06/29/21, 
        # Paul Rudnick said the phospho study was broken. Whenever the phospho data is fixed, this 
        # code should work to drop normal aliquots and average replicates. It would be a good idea to check when 
        # the new data is ready though. 
        # Phosphoproteomics
        phos = self._data["phosphoproteomics"]
        phos = phos.drop(drop_normals, level = 'aliquot_submitter_id') # drop normal aliquots (QC issues)
        phos = phos.rename(index={'604':'CPT000814'}) # use the aliquot for 604
        phos.index = phos.index.droplevel('aliquot_submitter_id')
        phos = average_replicates(phos, id_list = replicates) # average replicates
        phos = map_database_to_gene_pdc(phos, 'refseq') # map refseq IDs to gene names
        self._data["phosphoproteomics"] = phos
        
        # Acetylproteomics
        acetyl = self._data["acetylproteomics"]
        acetyl = acetyl.drop(drop_normals, level = 'aliquot_submitter_id') # drop normal aliquots (QC issues)
        acetyl = acetyl.rename(index={'604':'CPT000814'}) # use the aliquot for 604
        acetyl.index = acetyl.index.droplevel('aliquot_submitter_id') # drop aliquots
        acetyl = average_replicates(acetyl, id_list = replicates) # average replicates
        acetyl = map_database_to_gene_pdc(acetyl, 'refseq') # map refseq IDs to gene names
        self._data["acetylproteomics"] = acetyl

        
        # Sort rows (tumor first then normal) and columns by first level (protein/gene name)
        self._data = sort_all_rows_pancan(self._data)  


        print(" " * len(formatting_msg), end='\r') # Erase the formatting message



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import os
import warnings
import datetime
import logging
from gtfparse import read_gtf

from cptac.dataset import Dataset
from cptac.dataframe_tools import *
from cptac.exceptions import FailedReindexWarning, PublicationEmbargoWarning, ReindexMapError
from cptac.utils import get_boxnote_text
from .mssmclinical import MssmClinical


class WashuGbm(Dataset):

    def __init__(self, no_internet, version):
        """Load all of the bcmbrca dataframes as values in the self._data dict variable, with names as keys, and format them properly.

        Parameters:
        version (str, optional): The version number to load, or the string "latest" to just load the latest building. Default is "latest".
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """
        #ignore logging messages
        logger = logging.getLogger()
        logger.setLevel(logging.CRITICAL)

        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        # This keeps a record of all versions that the code is equipped to handle. That way, if there's a new data release but they didn't update their package, it won't try to parse the new data version it isn't equipped to handle.
        valid_versions = ["1.0"]

        data_files = {
            "1.0": [
                "GBM_discovery.dnp.annotated.exonic.maf.gz",
                "GBM_tumor_RNA-Seq_Expr_WashU_FPKM.tsv.gz",
                "GBM_mature_miRNA_combined.tsv",
                "GBM_precursor_miRNA_combined.tsv",
                "GBM_total_miRNA_combined.tsv",
                "GBM_xCell.txt",
                "CIBERSORT.Output_Abs_GBM.txt",
                "gencode.v22.annotation.gtf.gz",
                "GBM.gene_level.from_seg.filtered.tsv", 
                "CPTAC_pancan_RNA_tumor_purity_ESTIMATE_WashU.tsv.gz",
                "README_miRNA",
                "README_CIBERSORT",
                "README_xCell",
                "README_somatic_mutation_WXS",
                "README_gene_expression",
                "README.boxnote",
                "README_ESTIMATE_WashU"
                
            ]
        }

        # Call the parent class __init__ function
        super().__init__(cancer_type="washugbm", version=version, valid_versions=valid_versions, data_files=data_files, no_internet=no_internet)
        
        # get clinical df (used to slice out cancer specific patient_IDs in tumor_purity file)
        mssmclin = MssmClinical(no_internet=no_internet, version=version, filter_type='pancangbm') #_get_version - pancandataset
        clinical_df = mssmclin.get_clinical()
        

        # Load the data into dataframes in the self._data dict
        loading_msg = f"Loading {self.get_cancer_type()} v{self.version()}"
        for file_path in self._data_files_paths: # Loops through files variable

            # Print a loading message. We add a dot every time, so the user knows it's not frozen.
            loading_msg = loading_msg + "."
            print(loading_msg, end='\r')

            path_elements = file_path.split(os.sep) # Get a list of the levels of the path
            file_name = path_elements[-1] # The last element will be the name of the file. We'll use this to identify files for parsing in the if/elif statements below

            

            if file_name == "GBM_discovery.dnp.annotated.exonic.maf.gz": # Note that we use the "file_name" variable to identify files. That way we don't have to use the whole path.
                df = pd.read_csv(file_path, sep='\t')    
                df = pd.read_csv(file_path, sep='\t')    
                df['Patient_ID'] = df.loc[:, 'Tumor_Sample_Barcode']
                df = df.rename(columns={
                         "Hugo_Symbol":"Gene",
                         "Gene":"Gene_Database_ID",
                         "Variant_Classification":"Mutation",
                         "HGVSp_Short":"Location"})

                df = df.set_index("Patient_ID")
                df = df[ ['Gene'] + ["Mutation"] + ["Location"] + [ col for col in df.columns if col not in ["Gene","Mutation","Location"] ] ]
                df.index = df.index.str.replace(r"_T", "", regex=True)     
                self._data["somatic_mutation"] = df
                  
            if file_name == "GBM_tumor_RNA-Seq_Expr_WashU_FPKM.tsv.gz":
                df = pd.read_csv(file_path, sep="\t")
                df = df.rename(columns={"gene_name": "Name","gene_id": "Database_ID"})
                df = df.set_index(["Name", "Database_ID"])
                df = df.sort_index()
                df = df.T
                df.index.name = "Patient_ID"
                #remove label for tumor samples. All samples are tumors 
                df.index = df.index.str.replace(r"-T", "", regex=True) 
                self._data["transcriptomics"] = df
                
               
            elif 'miRNA_combined' in file_name:
                miRNA_type = file_name.split('_')[1] # get type of miRNA data (precursor, mature, or total)
                if miRNA_type == 'mature':
                    df = pd.read_csv(file_path, delimiter = '\t', index_col = ['Name', 'ID','Alias', 'Derives_from'])
                else:
                    df = pd.read_csv(file_path, delimiter = '\t', index_col = ['Name', 'ID','Alias'])
                df = df.transpose()
                df = average_replicates(df, common = '\.\d$') # average duplicate for C3N-02788
                df.index = df.index.str.replace('\.T$','', regex = True)
                df.index = df.index.str.replace('\.A$','.N', regex = True)
                df.index.name = 'Patient_ID'                
                # Sort
                normal = df.loc[df.index.str.contains('\.N$', regex =True)]
                normal = normal.sort_values(by=["Patient_ID"])
                tumor = df.loc[~ df.index.str.contains('\.N$', regex =True)]
                tumor = tumor.sort_values(by=["Patient_ID"])
                all_df = tumor.append(normal)
                self._data[miRNA_type+'_miRNA'] = all_df
                
            elif file_name == "GBM_xCell.txt":
                df = pd.read_csv(file_path, sep = '\t', index_col = 0) 
                df = df.transpose()
                df.columns.name = 'Name'
                df.index.name = 'Patient_ID'
                df.index = df.index.str.replace(r'-T$', '', regex=True) # remove label for tumor samples
                df.index = df.index.str.replace(r'-A$', '.N', regex=True) # change label for normal samples
                self._data["xcell"] = df
                
            elif file_name == "CIBERSORT.Output_Abs_GBM.txt":
                df = pd.read_csv(file_path, sep = '\t', index_col = 0) 
                df.index.name = 'Patient_ID'
                df.columns.name = 'Name'
                df.index = df.index.str.replace(r'-T$', '', regex=True) 
                df.index = df.index.str.replace(r'-A$', '.N', regex=True)
                self._data["cibersort"] = df
                
            elif file_name == "GBM.gene_level.from_seg.filtered.tsv":
                df = pd.read_csv(file_path, sep="\t")
                df = df.rename(columns={"Gene": "Name"})
                df = df.set_index("Name")
                self._data["CNV"] = df
                
            elif file_name == "gencode.v22.annotation.gtf.gz":
                df = read_gtf(file_path)
                df = df[["gene_name","gene_id"]]
                df = df.drop_duplicates()
                df = df.rename(columns={"gene_name": "Name","gene_id": "Database_ID"})
                df = df.set_index("Name")
                self._helper_tables["CNV_gene_ids"] = df
                
            elif file_name == "CPTAC_pancan_RNA_tumor_purity_ESTIMATE_WashU.tsv.gz":
                df = pd.read_csv(file_path, sep = "\t", na_values = 'NA')
                df.Sample_ID = df.Sample_ID.str.replace(r'-T', '', regex=True) # only tumor samples in file
                df = df.set_index('Sample_ID') 
                df.index.name = 'Patient_ID' 
                # Use list of patient_ids to slice out cancers                
                patient_ids = clinical_df.index.to_list()
                df = df.loc[df.index.isin(patient_ids)]                
                self._data["tumor_purity"] = df
                
            elif file_name == "README_miRNA":
                with open(file_path, 'r') as reader:
                    self._readme_files["readme_miRNA"] = reader.read()
                    
            elif file_name == "README_CIBERSORT":
                with open(file_path, 'r') as reader:
                    self._readme_files["readme_cibersort"] = reader.read()
                    
            elif file_name == "README_xCell":
                with open(file_path, 'r') as reader:
                    self._readme_files["readme_xcell"] = reader.read()
            
            elif file_name == "README_somatic_mutation_WXS":
                with open(file_path, 'r') as reader:
                    self._readme_files["readme_somatic_mutation"] = reader.read()
                    
            elif file_name == "README_gene_expression":
                with open(file_path, 'r') as reader:
                    self._readme_files["readme_transcriptomics"] = reader.read()
               
            elif file_name == "README.boxnote":
                self._readme_files["readme_cnv"] = get_boxnote_text(file_path)
            
            elif file_name == "README_ESTIMATE_WashU":
                with open(file_path, 'r') as reader:
                    self._readme_files["readme_tumor_purity"] = reader.read()
                
        print(' ' * len(loading_msg), end='\r') # Erase the loading message
        formatting_msg = f"Formatting {self.get_cancer_type()} dataframes..."
        print(formatting_msg, end='\r')
        
        # CNV
        cnv = self._data["CNV"]
        gene_ids = self._helper_tables["CNV_gene_ids"]
        df = cnv.join(gene_ids,how = "left") #merge in gene_ids 
        df = df.reset_index()
        df = df.set_index(["Name", "Database_ID"]) #create multi-index
        df = df.T
        df.index.name = 'Patient_ID'
        self._data["CNV"] = df
        
        self._data = sort_all_rows_pancan(self._data) # Sort IDs (tumor first then normal)


        print(" " * len(formatting_msg), end='\r') # Erase the formatting message



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import os
import warnings

from .dataset import Dataset
from .dataframe_tools import *
from .exceptions import FailedReindexWarning, ReindexMapError

class Ccrcc(Dataset):

    def __init__(self, version="latest", no_internet=False):
        """Load all of the ccrcc dataframes as values in the self._data dict variable, with names as keys, and format them properly.

        Parameters:
        version (str, optional): The version number to load, or the string "latest" to just load the latest building. Default is "latest".
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """

        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        valid_versions = ["0.0", "0.1", "0.1.1"] # This keeps a record of all versions that the code is equipped to handle. That way, if there's a new data release but they didn't update their package, it won't try to parse the new data version it isn't equipped to handle.

        data_files = {
            "0.0": [
                "6_CPTAC3_CCRCC_Phospho_abundance_gene_protNorm=2_CB_imputed.tsv.gz",
                "6_CPTAC3_CCRCC_Phospho_abundance_phosphosite_protNorm=2_CB.tsv.gz",
                "6_CPTAC3_CCRCC_Whole_abundance_protein_pep=unique_protNorm=2_CB.tsv.gz",
                "Clinical Table S1.xlsx",
                "ccrcc.somatic.consensus.gdc.umichigan.wu.112918.maf.gz",
                "ccrccMethylGeneLevelByMean.txt.gz",
                "cptac-metadata.xls.gz",
                "kirc_wgs_cnv_gene.csv.gz",
                "RNA_Normal_Tumor_185_samples.tsv.gz",
                "S044_CPTAC_CCRCC_Discovery_Cohort_Clinical_Data_r3_Mar2019.xlsx"],
            "0.1": [
                "6_CPTAC3_CCRCC_Phospho_abundance_gene_protNorm=2_CB_imputed.tsv.gz",
                "6_CPTAC3_CCRCC_Phospho_abundance_phosphosite_protNorm=2_CB.tsv.gz",
                "6_CPTAC3_CCRCC_Whole_abundance_protein_pep=unique_protNorm=2_CB.tsv.gz",
                "ccrccMethylGeneLevelByMean.txt.gz",
                "ccrcc.somatic.consensus.gdc.umichigan.wu.112918.maf.gz",
                "Clinical Table S1.xlsx",
                "cptac-metadata.xls.gz",
                "kirc_wgs_cnv_gene.csv.gz",
                "RNA_Normal_Tumor_185_samples.tsv.gz",
                "S044_CPTAC_CCRCC_Discovery_Cohort_Clinical_Data_r3_Mar2019.xlsx",
                "Table S7.xlsx"],
            "0.1.1": [
                "6_CPTAC3_CCRCC_Phospho_abundance_gene_protNorm=2_CB_imputed.tsv.gz",
                "6_CPTAC3_CCRCC_Phospho_abundance_phosphosite_protNorm=2_CB.tsv.gz",
                "6_CPTAC3_CCRCC_Whole_abundance_protein_pep=unique_protNorm=2_CB.tsv.gz",
                "CCRCC_followup_9_12.xlsx",
                "ccrccMethylGeneLevelByMean.txt.gz",
                "ccrcc.somatic.consensus.gdc.umichigan.wu.112918.maf.gz",
                "Clinical Table S1.xlsx",
                "cptac-metadata.xls.gz",
                "kirc_wgs_cnv_gene.csv.gz",
                "RNA_Normal_Tumor_185_samples.tsv.gz",
                "S044_CPTAC_CCRCC_Discovery_Cohort_Clinical_Data_r3_Mar2019.xlsx",
                "Table S7.xlsx"],
        }

        super().__init__(cancer_type="ccrcc", version=version, valid_versions=valid_versions, data_files=data_files, no_internet=no_internet)

        # We're going to need to drop the samples below from a couple dataframes
        nci_labels = ["NCI7-1", "NCI7-2", "NCI7-3", "NCI7-4", "NCI7-5"]
        nci_dotted_labels = [label.replace("-", ".") for label in nci_labels]
        qc_labels = ["QC1", "QC2", "QC3", "QC4", "QC5", "QC6", "QC7", "QC8"]

        # We have multiple clinical files, so we'll load those into a separate dict, and combine them later
        clinical_dfs = {}

        # Load the data into dataframes in the self._data dict
        loading_msg = f"Loading {self.get_cancer_type()} v{self.version()}"
        for file_path in self._data_files_paths: # Loops through files variable

            # Print a loading message. We add a dot every time, so the user knows it's not frozen.
            loading_msg = loading_msg + "."
            print(loading_msg, end='\r')

            path_elements = file_path.split(os.sep) # Get a list of the levels of the path
            file_name = path_elements[-1] # The last element will be the name of the file
            df_name = file_name.split(".")[0] # Our dataframe name will be the first section of file name (i.e. proteomics.txt.gz becomes proteomics)

            if file_name == "6_CPTAC3_CCRCC_Phospho_abundance_gene_protNorm=2_CB_imputed.tsv.gz":
                df = pd.read_csv(file_path, sep='\t', index_col=0)
                ref_intensities = df["ReferenceIntensity"] # Copy this out, so we can subtract the reference intensities later
                df = df.drop(columns=["NumberPSM", "Proteins", "ReferenceIntensity"] + nci_dotted_labels + qc_labels)
                df = df.subtract(ref_intensities, axis="index") # Subtract reference intensities from all the values, to get ratios
                df = df.transpose()
                self._data["phosphoproteomics_gene"] = df

            elif file_name == "6_CPTAC3_CCRCC_Phospho_abundance_phosphosite_protNorm=2_CB.tsv.gz":
                df = pd.read_csv(file_path, sep='\t')

                # Drop unlocalized sites
                unlocalized_sites = (df["Index"].str.rsplit("_", n=1, expand=True)[1] == '0') 
                df = df[~unlocalized_sites]

                # Drop unwanted samples
                df = df.drop(columns=nci_labels + qc_labels)

                # Subtract reference intensities from numerical data columns, to get ratios
                df = df.rename(columns={"Gene": "Name"})
                metadata_cols = ["Index", "Name", "Peptide", "ReferenceIntensity"]
                metadata = df[metadata_cols] # Extract these for later
                df = df.drop(columns=metadata_cols) # Get the df to contain just the numerical data columns
                ref_intensities = metadata["ReferenceIntensity"]
                df = df.subtract(ref_intensities, axis="index") # Subtract reference intensities
                df = metadata.join(df, how="outer") # Put the metadata columns back in
                df = df.drop(columns="ReferenceIntensity") # Don't need this anymore

                # Parse a few columns out of the "Index" column that we'll need for our multiindex
                split_ids = df["Index"].str.split('_', expand=True)
                df = df.drop(columns="Index")
                sites = split_ids.iloc[:, -1]
                database_ids = split_ids[0].str.cat(split_ids[1], sep='_')
                df = df.assign(**{"Site": sites, "Database_ID": database_ids})

                # Some rows have at least one localized phosphorylation site, but also have other phosphorylations that aren't localized. We'll drop those rows, if their localized sites are duplicated in another row, to avoid creating duplicates, because we only preserve information about the localized sites in a given row. However, if the localized sites aren't duplicated in another row, we'll keep the row.
                unlocalized_to_drop = df.index[~split_ids[4].eq(split_ids[5]) & df.duplicated(["Name", "Site", "Peptide", "Database_ID"], keep=False)] # Column 4 of the split "Index" column is number of phosphorylations detected, and column 5 is number of phosphorylations localized, so if the two values aren't equal, the row has at least one unlocalized site
                df = df.drop(index=unlocalized_to_drop)

                # Give it a multiindex
                df = df.set_index(["Name", "Site", "Peptide", "Database_ID"]) # This will create a multiindex from these columns, in this order.
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                self._data["phosphoproteomics"] = df
            
            elif file_name == "6_CPTAC3_CCRCC_Whole_abundance_protein_pep=unique_protNorm=2_CB.tsv.gz":
                df = pd.read_csv(file_path, sep='\t')
                df = df.rename(columns={"Proteins": "Name", "Index": "Database_ID"})
                df = df.set_index(["Name", "Database_ID"])
                ref_intensities = df["ReferenceIntensity"] # Copy this out, so we can subtract the reference intensities later
                df = df.drop(columns=["NumberPSM", "ReferenceIntensity"] + nci_labels + qc_labels)
                df = df.subtract(ref_intensities, axis="index") # Subtract reference intensities from all the values, to get ratios
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                self._data["proteomics"] = df
            
            elif file_name == "Clinical Table S1.xlsx":
                df = pd.\
                    read_excel(file_path, sheet_name="ccrcc_clinical_characteristics", index_col=0).\
                    dropna(how="all", axis=0).\
                    dropna(how="all", axis=1)
                df.index.name = "Patient_ID" # The index is currently "case_id", but we call that "Patient_ID"
                clinical_dfs["authoritative_clinical"] = df
            
            elif file_name == "ccrcc.somatic.consensus.gdc.umichigan.wu.112918.maf.gz":
                df = pd.read_csv(file_path, sep='\t', dtype={"PUBMED":object}) # "PUBMED" column has mixed types, so we specify object as the dtype to prevent a warning from printing. We don't actually use the column, so that's all we need to do.
                split_barcode = df["Tumor_Sample_Barcode"].str.split("_", n=1, expand=True) # The first part of the barcode is the patient id, which we need want to make the index
                df["Tumor_Sample_Barcode"] = split_barcode[0]
                df = df[["Tumor_Sample_Barcode","Hugo_Symbol","Variant_Classification","HGVSp_Short"]]
                df = df.rename(columns={"Tumor_Sample_Barcode":"Patient_ID","Hugo_Symbol":"Gene","Variant_Classification":"Mutation","HGVSp_Short":"Location"})                
                df = df.sort_values(by=["Patient_ID", "Gene"])
                df = df.set_index("Patient_ID")
                self._data["somatic_mutation"] = df

            elif file_name == "ccrccMethylGeneLevelByMean.txt.gz":
                df = pd.read_csv(file_path, sep='\t')
                df = df.sort_index()
                df = df.transpose()
                df.index.name = "Patient_ID"
                self._data["methylation"] = df

            elif file_name == "cptac-metadata.xls.gz":
                df = pd.read_csv(file_path, index_col=0)
                df = df.drop(index=["pooled sample"] + nci_labels + qc_labels) # Drop the pooled samples in addition to other samples to exclude
                clinical_dfs["metadata_and_keys"] = df
            
            elif file_name == "kirc_wgs_cnv_gene.csv.gz":
                df = pd.read_csv(file_path)
                df = df.rename(columns={"gene_name": "Name", "gene_id": "Database_ID"})
                df = df.set_index(["Name", "Database_ID"])
                df = df.sort_index()
                df = df.transpose()

                # The dataframe contains 4 rows for each sample: lr.loc_***,  lr.seg_***, mzd.loc_***, or  mzd.seg_*** where *** is the 
                # Patient_ID. "lr" stands for log ratio, and "mzd" stands for Mean Zygosity Deviation. “lr.loc” is based on average of lr 
                # of probes belonging to each gene. “lr.seg” is based on segmented CNV result (i.e. first performed segmentation based on 
                # probe level data, and then use the lr of the representative segment of each gene as the gene-level lr). They used the
                # lr.seg values in the paper, so we'll use those.
                df = df.drop(index=df[~df.index.str.startswith("lr.seg")].index)

                # Parse a Patient_ID index out of the current index
                barcode_col = df.index.to_series()
                split_barcode = barcode_col.str.split("_", n=1, expand=True) # The second part of the barcode is the patient id, which we want to make the index
                df.index = pd.Index(split_barcode[1])
                df.index.name = "Patient_ID"

                df = df.sort_index()
                self._data["CNV"] = df

            elif file_name == "RNA_Normal_Tumor_185_samples.tsv.gz":
                df = pd.read_csv(file_path, sep='\t')
                df = df.sort_index()
                df = df.transpose()

                # There are a couple duplicate column headers, but they're full of just zeros. We'll drop them.
                # You can do this with a one liner: df =  df.loc[:, ~df.columns.duplicated(keep=False) | (df != 0).any(axis=0)]
                # But the one liner is about 100 times slower.
                dups = df.loc[:, df.columns.duplicated(keep=False)] # Select all the columns with duplicated headers
                dups = dups.loc[:, (dups != 0).any(axis=0)] # Get only the columns that aren't all zeros
                df = df.loc[:, ~df.columns.duplicated(keep=False)] # Get rid of the duplicate columns from the original dataframe
                df = df.join(dups, how="outer") # Sub in our un-duplicated selections
                df = df.sort_index(axis="columns") # Get all the column names in order again

                self._data["transcriptomics"] = df
            
            elif file_name == "S044_CPTAC_CCRCC_Discovery_Cohort_Clinical_Data_r3_Mar2019.xlsx":

                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", "Unknown extension is not supported and will be removed") # This warning is just due to some formatting on the spreadsheet. We don't need to worry about it.
                    clinical_sheets = pd.read_excel(file_path,
                        sheet_name=['Patient_Clinical_Attributes', 'Other_Medical_Information', 'Specimen_Attributes'],
                        index_col=0,
                        usecols=lambda x: x != "tumor_code") # Don't load the tumor_code column in any of them--it's just "CCRCC" for every row.

                for sheet, df in clinical_sheets.items(): # Put them in the clinical_dfs dictionary for processing later
                    df.index.name = "Patient_ID" # The indices are currently "case_id", but we call that "Patient_ID"
                    clinical_dfs[sheet] = df

            elif file_name == "Table S7.xlsx":
                immune_groups = pd.\
                    read_excel(
                        file_path, 
                        sheet_name="xCell Signatures", 
                        index_col = 0, 
                        header=None,
                        skiprows=range(0, 3)
                    ).\
                    transpose()
                immune_groups = immune_groups[["Samples", "Immune Group"]] # We only need these columns
                immune_groups = immune_groups.set_index("Samples")

            elif file_name == 'CCRCC_followup_9_12.xlsx' and self._version == "0.1.1":
                df = pd.read_excel(file_path)

                # Replace redundant values for "not reported" with NaN
                nan_equivalents = ['Not Reported/ Unknown', 'Reported/ Unknown', 'Not Applicable', 'na',
                    'unknown', 'Not Performed', 'Unknown tumor status', 'Unknown', ' Unknown', 'Unknown ',
                    'Unknown Tumor Status', 'Not specified']

                df = df.replace(nan_equivalents, np.nan)

                # Replace redundanct values in "Cause of Death" column
                disease_prog_equivalents = ['Metastatic Renal Cell Carcinoma', 'Tumor progression',
                    'Progression of disease', 'Progression of disease ', 'Tumor', 'Disease progression',
                    'Progressive Disease', 'Disease progression', 'disease progression ', 'main disease ']

                df['Cause of Death'] = df['Cause of Death'].replace(disease_prog_equivalents, 'Disease progression')

                # Rename, set, and sort by index
                df = df.rename(columns={"Case ID": "Patient_ID"})
                df = df.set_index("Patient_ID")
                df = df.sort_index()

                self._data["followup"] = df

        print(' ' * len(loading_msg), end='\r') # Erase the loading message
        formatting_msg = "Formatting dataframes..."
        print(formatting_msg, end='\r')

        # Process and combine the multiple clinical dataframes
        clinical = clinical_dfs["metadata_and_keys"] # We'll start with this dataframe, and add the others to it.
        clinical.index = clinical.index.where(~(clinical["Type"] == "Normal"), 'N' + clinical.index) # Prepend an 'N' to the patient IDs of normal samples
        clinical.index.name = "Patient_ID" # Name the index Patient_ID, since the index values are in fact the patient IDs (also called case IDs)
        clinical = clinical.sort_index()
        clinical = clinical.rename(columns={"Type": "Sample_Tumor_Normal"}) # This matches the other datasets.
        clinical = clinical.join(clinical_dfs["Patient_Clinical_Attributes"], how="outer")

        auth_clin = clinical_dfs["authoritative_clinical"] # This table has more authoritative values for histologic type, so we'll replace that column
        clinical["histologic_type"] = auth_clin["Histologic_Type"]

        specimen_attributes = clinical_dfs["Specimen_Attributes"] # Before we can join in this dataframe, we need to prepend "N" to the normal samples' index values
        specimen_attributes.index = np.where(specimen_attributes["tissue_type"] == "normal", 'N' + specimen_attributes.index, specimen_attributes.index)
        specimen_attributes.index.name = "Patient_ID" # Name the index Patient_ID, since the index values are in fact the patient IDs (also called case IDs)
        clinical = clinical.join(specimen_attributes, how="outer") # Join in our nicely reindexed dataframe

        medical_info = clinical_dfs["Other_Medical_Information"] # We're going to join one column from this to the clinical dataframe, and later take some other columns to make a medical_history dataframe
        medication_col = medical_info["medication_name"]
        medication_col = medication_col.str.replace("|", ",", regex=False)
        clinical = clinical.assign(patient_medications=medication_col) # Add it in as a new "patient_medications" column
        clinical = clinical.drop(columns=["MS.Directory.Name", "Batch", "Data.Set", "TMT.Channel", "Mass.Spectrometer", "Mass.Spectrometer.Operator", "Set.A", "Set.B", "tissue_type"]) # tissue_type column is a duplicate of Sample_Tumor_Normal

        # If this if version 0.1 or greater, join the immune group data into the clinical dataframe
        if self._version == "0.1":
            clinical = clinical.join(immune_groups, on="Specimen.Label", how="outer")

        # Save this final amalgamation of all the clinical dataframes
        self._data["clinical"] = clinical 

        # Create the medical_history dataframe
        medical_info = clinical_dfs["Other_Medical_Information"]
        medical_history_unparsed = medical_info[["medical_condition", "medical_condition_other", "condition_year_of_onset", "history_of_medical_treatment", "medical_history_source"]]
        medical_history_unparsed = medical_history_unparsed.dropna(axis="index", how="all")

        first_time = True # We'll use this to get special behavior on our first iteration
        for col_name in medical_history_unparsed.columns:
            med_hist_col = medical_history_unparsed[col_name].str.split("|", expand=True) # Cells have mutliple values separated by "|", so we'll splits those into individual columns
            med_hist_col = med_hist_col.reset_index() # Make the sample ids a column so we can use them in melt
            med_hist_col = med_hist_col.melt(id_vars="Patient_ID") # Use each row's patient id as its index, then combine everything else into one column
            if first_time:
                med_hist_col = med_hist_col.drop(columns="variable") # Had the old column headers
            else:
                med_hist_col = med_hist_col.drop(columns=["Patient_ID", "variable"]) # Also dropping patient ids so we're not trying to join dfs with duplicate columns at the end of the loop
            med_hist_col = med_hist_col.rename(columns={"value": col_name})
            med_hist_col = med_hist_col.dropna()
            if first_time:
                medical_history_parsed = med_hist_col 
            else:
                medical_history_parsed = medical_history_parsed.join(med_hist_col, how="outer")
            first_time = False

        med_cond = medical_history_parsed["medical_condition"] # Where medical_condition column is "Other, specify", we're going to fill in value from medical_condition_other column
        med_cond = med_cond.where(med_cond != "Other, specify", medical_history_parsed["medical_condition_other"])
        medical_history_parsed["medical_condition"] = med_cond
        medical_history_parsed = medical_history_parsed.drop(columns="medical_condition_other") # Since we copied all of the useful information into the medical_condition column

        medical_history_parsed = medical_history_parsed.set_index("Patient_ID")
        medical_history_parsed = medical_history_parsed.sort_index()
        self._data["medical_history"] = medical_history_parsed

        # Use the RNA.ID column from clinical dataframe to reindex transcriptomics dataframe with patient ids
        tran = self._data["transcriptomics"]
        try:
            tran_map = get_reindex_map(clinical["RNA.ID"])
            tran_reindexed = reindex_dataframe(tran, tran_map, new_index_name="Patient_ID", keep_old=False)
        except ReindexMapError:
            del self._data["transcriptomics"]
            warnings.warn("Error mapping sample ids in transcriptomics dataframe. At least one RNA.ID did not have a corresponding Patient_ID mapped in the clinical dataframe. transcriptomics dataframe not loaded.", FailedReindexWarning, stacklevel=2)
        else:
            self._data["transcriptomics"] = tran_reindexed

        # Use the Specimen.Label columns from clinical dataframe to reindex the proteomics, phosphoproteomics, and phosphoproteomics_gene dataframes with patient ids
        specimen_indexed_dfs = ["proteomics", "phosphoproteomics", "phosphoproteomics_gene"]
        try:
            specimen_label_map = get_reindex_map(clinical["Specimen.Label"])
        except ReindexMapError:
            for df_name in specimen_indexed_dfs:
                del self._data[df_name]
            warnings.warn(f"Error mapping sample ids in these dataframes: {' '.join(df for df in specimen_indexed_dfs)}. Specimen.Label mapping in clinical dataframe was not one-to-one. Dataframes not loaded.", FailedReindexWarning, stacklevel=2)
        else:
            for df_name in specimen_indexed_dfs:
                df = self._data[df_name]
                try:
                    df_reindexed = reindex_dataframe(df, specimen_label_map, new_index_name="Patient_ID", keep_old=False)
                except ReindexMapError as error:
                    del self._data[df_name]
                    warnings.warn(f"Error mapping sample ids in {df_name} dataframe. RNA.ID {str(error)} did not have a corresponding Patient_ID mapped in the clinical dataframe. {df_name} dataframe not loaded.", FailedReindexWarning, stacklevel=2)
                else:
                    self._data[df_name] = df_reindexed

        # Now that we've used the RNA.ID and Specimen.Label columns to reindex the dataframes that needed it, we can drop them from the clinical dataframe
        clinical = clinical.drop(columns=["Specimen.Label", "RNA.ID"])

        # Replace the clinical dataframe in the data dictionary with our new and improved version!
        self._data['clinical'] = clinical

        # There are 28 samples in the CNV table, 26 of which are also in the methylation table, that aren't in any other tables, and were actually excluded from analysis. We're going to drop them.
        # All samples used for analysis should be in the proteomics dataframe, so we'll use that test to determine which samples to drop from CNV and methylation.
        prot = self._data["proteomics"]

        cnv = self._data["CNV"]
        to_drop = cnv.index.difference(prot.index)
        cnv = cnv.drop(index=to_drop)
        self._data["CNV"] = cnv

        methylation = self._data["methylation"]
        to_drop = methylation.index.difference(prot.index)
        methylation = methylation.drop(index=to_drop)
        self._data["methylation"] = methylation

        # Get a union of all dataframes' indices, with duplicates removed
        master_index = unionize_indices(self._data, exclude="followup")

        # Use the master index to reindex the clinical dataframe, so the clinical dataframe has a record of every sample in the dataset. Rows that didn't exist before (such as the rows for normal samples) are filled with NaN.
        clinical = self._data["clinical"]
        clinical = clinical.reindex(master_index)
        self._data['clinical'] = clinical

        # Edit the format of the Patient_IDs to have normal samples marked the same way as in other datasets. Currently, normal patient IDs have an "N" prepended. We're going to erase that and append a ".N"
        self._data = reformat_normal_patient_ids(self._data, existing_identifier="N", existing_identifier_location="start")

        # Call function from dataframe_tools.py to sort all tables first by sample status, and then by the index
        self._data = sort_all_rows(self._data)

        # Call function from dataframe_tools.py to standardize the names of the index and column axes
        self._data = standardize_axes_names(self._data)

        print(" " * len(formatting_msg), end='\r') # Erase the formatting message

    def how_to_cite(self):
        return super().how_to_cite(cancer_type='clear cell renal cell carcinoma (kidney)', pmid=31675502)


#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import os
import warnings
import datetime
from .dataset import Dataset
from .dataframe_tools import *
from .exceptions import FailedReindexWarning, PublicationEmbargoWarning, ReindexMapError

class Luad(Dataset):

    def __init__(self, version="latest", no_internet=False):
        """Load all of the luad dataframes as values in the self._data dict variable, with names as keys, and format them properly.

        Parameters:
        version (str, optional): The version number to load, or the string "latest" to just load the latest building. Default is "latest".
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """

        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        valid_versions = ["2.0", "3.1", "3.1.1"] # This keeps a record of all versions that the code is equipped to handle. That way, if there's a new data release but they didn't update their package, it won't try to parse the new data version it isn't equipped to handle.

        data_files = {
            "2.0": [
                "luad-v2.0-cnv-gene-LR.gct.gz",
                "luad-v2.0-phosphoproteome-ratio-norm-NArm.gct.gz",
                "luad-v2.0-proteome-ratio-norm-NArm.gct.gz",
                "luad-v2.0-rnaseq-circ-rna.csv.gz",
                "luad-v2.0-rnaseq-prot-uq-rpkm-log2-NArm-row-norm.gct.gz",
                "luad-v2.0-sample-annotation.csv.gz"],
            "3.1": [
                "LUAD_followup_9_12.xlsx",
                "luad-v3.0-rnaseq-circ-rna.csv.gz",
                "luad-v3.0-rnaseq-gene-fusions.csv.gz",
                "luad-v3.0-wxs-somatic.luad.v1.4.20190517.maf.gz",
                "luad-v3.1-acetylome-ratio-norm-NArm.gct.gz",
                "luad-v3.1-cnv-gene-LR.gct.gz",
                "luad-v3.1-mirna-mature-tpm-log2.gct.gz",
                "luad-v3.1-phosphoproteome-ratio-norm-NArm.gct.gz",
                "luad-v3.1-proteome-ratio-norm-NArm.gct.gz",
                "luad-v3.1-rnaseq-linc-uq-rpkm-log2-NArm.gct.gz",
                "luad-v3.1-rnaseq-prot-uq-rpkm-log2-NArm.gct.gz",
                "luad-v3.1-sample-annotation.csv.gz"],
            "3.1.1": [
                "LUAD_followup_9_12.xlsx",
                "luad-v3.0-rnaseq-circ-rna_parsed.tsv.gz",
                "luad-v3.0-rnaseq-gene-fusions.csv.gz",
                "luad-v3.0-wxs-somatic.luad.v1.4.20190517.maf.gz",
                "luad-v3.1-acetylome-ratio-norm-NArm.gct.gz",
                "luad-v3.1-cnv-gene-LR.gct.gz",
                "luad-v3.1-mirna-mature-tpm-log2.gct.gz",
                "luad-v3.1-phosphoproteome-ratio-norm-NArm.gct.gz",
                "luad-v3.1-proteome-ratio-norm-NArm.gct.gz",
                "luad-v3.1-rnaseq-linc-uq-rpkm-log2-NArm.gct.gz",
                "luad-v3.1-rnaseq-prot-uq-rpkm-log2-NArm.gct.gz",
                "luad-v3.1-sample-annotation.csv.gz"],
        }

        super().__init__(cancer_type="luad", version=version, valid_versions=valid_versions, data_files=data_files, no_internet=no_internet)

        # Load the data into dataframes in the self._data dict
        loading_msg = f"Loading {self.get_cancer_type()} v{self.version()}"
        for file_path in self._data_files_paths: # Loops through files variable

            # Print a loading message. We add a dot every time, so the user knows it's not frozen.
            loading_msg = loading_msg + "."
            print(loading_msg, end='\r')

            path_elements = file_path.split(os.sep) # Get a list of the levels of the path
            file_name = path_elements[-1] # The last element will be the name of the file
            df_name = file_name.split(".")[0] # Our dataframe name will be the first section of file name (i.e. proteomics.txt.gz becomes proteomics)

            if file_name == "luad-v3.0-rnaseq-gene-fusions.csv.gz":
                df = pd.read_csv(file_path)
                df = df.rename(columns={"Sample.ID": "Patient_ID"})
                df = df.set_index("Patient_ID")

                self._data["gene_fusion"] = df

            elif file_name == "luad-v3.0-wxs-somatic.luad.v1.4.20190517.maf.gz":
                df = pd.read_csv(file_path, sep='\t')
                df = df.rename(columns={"Sample.ID": "Patient_ID"})

                df = df[['Patient_ID','Hugo_Symbol','Variant_Classification','HGVSp_Short']]
                df = df.rename(columns={
                    "Hugo_Symbol":"Gene",
                    "Variant_Classification":"Mutation",
                    "HGVSp_Short":"Location"}) # Rename the columns we want to keep to the appropriate names

                df = df.sort_values(by=["Patient_ID", "Gene"])
                df = df.set_index("Patient_ID")

                self._data["somatic_mutation"] = df

            elif file_name == "luad-v3.1-acetylome-ratio-norm-NArm.gct.gz":
                df = pd.read_csv(file_path, sep="\t", skiprows=2, dtype=object)
                gene_filter = df['geneSymbol'] != 'na' #Drop rows of metadata
                df = df[gene_filter]

                # Prepare some columns we'll need later for the multiindex
                df["variableSites"] = df["variableSites"].str.replace(r"[a-z\s]", "", regex=True) # Get rid of all lowercase delimeters and whitespace in the sites
                df = df.rename(columns={
                    "geneSymbol": "Name",
                    "variableSites": "Site",
                    "sequence": "Peptide", # We take this instead of sequenceVML, to match the other datasets' format
                    "accession_numbers": "Database_ID" # We take all accession numbers they have, instead of the singular accession_number column
                    })


                # Some rows have at least one localized acetylation site, but also have other acetylations that aren't localized. We'll drop those rows, if their localized sites are duplicated in another row, to avoid creating duplicates, because we only preserve information about the localized sites in a given row. However, if the localized sites aren't duplicated in another row, we'll keep the row.
                localization = df["accessionNumber_VMsites_numVMsitesPresent_numVMsitesLocalizedBest_earliestVMsiteAA_latestVMsiteAA"].str.split("_", expand=True)
                unlocalized_to_drop = localization.index[~localization[3].eq(localization[4]) & df.duplicated(["Name", "Site", "Peptide", "Database_ID"], keep=False)] # Column 3 of the split localization data column is number of acetylation sites detected, and column 4 is number of acetylation sites localized, so if the two values aren't equal, the row has at least one unlocalized site
                df = df.drop(index=unlocalized_to_drop)

                # Give it a multiindex
                df = df.set_index(["Name", "Site", "Peptide", "Database_ID"])                

                cols_to_drop = [
                    "id",
                    "id.description",
                    "numColumnsVMsiteObserved",
                    "bestScore",
                    "bestDeltaForwardReverseScore",
                    "Best_scoreVML",
                    "sequenceVML",
                    "accessionNumber_VMsites_numVMsitesPresent_numVMsitesLocalizedBest_earliestVMsiteAA_latestVMsiteAA",
                    "protein_mw",
                    "species",
                    "speciesMulti",
                    "orfCategory",
                    "accession_number",
                    "protein_group_num",
                    "entry_name",
                    "GeneSymbol",
                    ]
                df = df.drop(columns=cols_to_drop)

                df = df.apply(pd.to_numeric)
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.index.name="Patient_ID"
                
                self._data["acetylproteomics"] = df

            elif file_name == "luad-v3.1-mirna-mature-tpm-log2.gct.gz":
                df = pd.read_csv(file_path, sep="\t", skiprows=2, dtype=object)

                # Filter out metadata rows
                gene_filter = df["Name"] != 'na' 
                df = df[gene_filter]

                df = df.drop(columns=['id', 'Alias', 'Name', 'Derives_from', 'Quantified.in.Percent.Samples'])
                df = df.set_index("ID")
                df = df.apply(pd.to_numeric)
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()

                self._data["miRNA"] = df

            elif file_name == "luad-v3.1-rnaseq-linc-uq-rpkm-log2-NArm.gct.gz":
                df = pd.read_csv(file_path, sep="\t", skiprows=2, dtype=object)

                # Filter out metadata rows
                gene_filter = df["geneSymbol"] != 'na' 
                df = df[gene_filter]

                # Filter out just lincRNA. Current it has a bunch of other RNA types too. We'll worry about them later.
                lincRNA_filter = df["gene_type"] == "lincRNA"
                df = df[lincRNA_filter]

                # Drop unneeded columns, set index, sort, transpose
                df = df.drop(columns=["id", "gene_id", "gene_type", "length"])
                df = df.rename(columns={"geneSymbol": "Name"})
                df = df.set_index("Name")
                df = df.apply(pd.to_numeric)
                df = df.sort_index()
                df = df.transpose()
                df.index.name = "Patient_ID"

                self._data["lincRNA"] = df

            elif file_name == "luad-v2.0-cnv-gene-LR.gct.gz" or file_name == "luad-v3.1-cnv-gene-LR.gct.gz":
                df = pd.read_csv(file_path, sep="\t", skiprows=2, dtype=object)

                # Filter out metadata rows
                if self._version == "2.0":
                    gene_filter = df['Description'] != 'na' 

                elif self._version in ["3.1", "3.1.1"]:
                    gene_filter = df['geneSymbol'] != 'na' 

                df = df[gene_filter]

                # Drop non-quantitative columns
                if self._version == "2.0":
                    cols_to_drop = ["GeneID","Description"]
                    
                elif self._version in ["3.1", "3.1.1"]:
                    cols_to_drop = ["id", "gene_id"]

                df = df.drop(columns=cols_to_drop)

                # Set index
                if self._version == "2.0":
                    df = df.set_index("id")

                elif self._version in ["3.1", "3.1.1"]:
                    df = df.set_index("geneSymbol")

                df = df.apply(pd.to_numeric)
                df = df.sort_index()
                df = df.transpose()

                df = df.sort_index()
                df.index.name="Patient_ID"

                self._data["CNV"] = df

            elif file_name == "luad-v2.0-phosphoproteome-ratio-norm-NArm.gct.gz" or file_name == "luad-v3.1-phosphoproteome-ratio-norm-NArm.gct.gz":
                df = pd.read_csv(file_path, sep="\t", skiprows=2, dtype=object)

                # Drop rows of metadata
                gene_filter = df['geneSymbol'] != 'na' 
                df = df[gene_filter]

                # Prepare some columns we'll need later for the multiindex
                df["variableSites"] = df["variableSites"].str.replace(r"[a-z\s]", "", regex=True) # Get rid of all lowercase delimeters and whitespace in the sites
                df = df.rename(columns={
                    "geneSymbol": "Name",
                    "variableSites": "Site",
                    "sequence": "Peptide", # We take this instead of sequenceVML, to match the other datasets' format
                    "accession_numbers": "Database_ID" # We take all accession numbers they have, instead of the singular accession_number column
                    })

                # Some rows have at least one localized phosphorylation site, but also have other phosphorylations that aren't localized. We'll drop those rows, if their localized sites are duplicated in another row, to avoid creating duplicates, because we only preserve information about the localized sites in a given row. However, if the localized sites aren't duplicated in another row, we'll keep the row.
                unlocalized_to_drop = df.index[~df['Best_numActualVMSites_sty'].eq(df['Best_numLocalizedVMsites_sty']) & df.duplicated(["Name", "Site", "Peptide", "Database_ID"], keep=False)] # Column 3 of the split "id" column is number of phosphorylations detected, and column 4 is number of phosphorylations localized, so if the two values aren't equal, the row has at least one unlocalized site
                df = df.drop(index=unlocalized_to_drop)

                # Give it a multiindex
                df = df.set_index(["Name", "Site", "Peptide", "Database_ID"])                

                # Drop non-quantitative columns
                if self._version == "2.0":
                    cols_to_drop = ['id', 'id.1', 'id.description', 'numColumnsVMsiteObserved', 'bestScore', 'bestDeltaForwardReverseScore',
                        'Best_scoreVML', 'Best_numActualVMSites_sty', 'Best_numLocalizedVMsites_sty', 'sequenceVML',
                        'accessionNumber_VMsites_numVMsitesPresent_numVMsitesLocalizedBest_earliestVMsiteAA_latestVMsiteAA', 'protein_mw', 'species',
                        'speciesMulti', 'orfCategory', 'accession_number', 'protein_group_num', 'entry_name', 'GeneSymbol']
                elif self._version in ["3.1", "3.1.1"]:
                    cols_to_drop = ["id", "id.description", "numColumnsVMsiteObserved", "bestScore", "bestDeltaForwardReverseScore",
                        "Best_scoreVML", "Best_numActualVMSites_sty", "Best_numLocalizedVMsites_sty", "sequenceVML",
                        "accessionNumber_VMsites_numVMsitesPresent_numVMsitesLocalizedBest_earliestVMsiteAA_latestVMsiteAA", "protein_mw", "species",
                        "speciesMulti", "orfCategory", "accession_number", "protein_group_num", "entry_name", "GeneSymbol"]
                
                df = df.drop(columns=cols_to_drop)

                # Format table
                df = df.apply(pd.to_numeric)
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.index.name="Patient_ID"

                self._data["phosphoproteomics"] = df

            elif file_name == "luad-v2.0-proteome-ratio-norm-NArm.gct.gz" or file_name == "luad-v3.1-proteome-ratio-norm-NArm.gct.gz":
                df = pd.read_csv(file_path, skiprows=2, sep='\t', dtype=object)

                # Filter out rows of metadata
                gene_filter = df['geneSymbol'] != 'na' 
                df = df[gene_filter]

                # Set multiindex
                df = df.rename(columns={"GeneSymbol": "Name", 'accession_numbers': "Database_ID"})
                df = df.set_index(["Name", "Database_ID"])

                # Drop non-quantitative columns
                if self._version == "2.0":
                    cols_to_drop = ['id', 'id.1', 'id.description', 'geneSymbol', 'numColumnsProteinObserved', 'numSpectraProteinObserved',
                        'protein_mw', 'percentCoverage', 'numPepsUnique', 'scoreUnique', 'species', 'orfCategory', 'accession_number',
                        'subgroupNum', 'entry_name']
                elif self._version in ["3.1", "3.1.1"]:
                    cols_to_drop = ["id", "id.description", "geneSymbol", "numColumnsProteinObserved", "numSpectraProteinObserved",
                        "protein_mw", "percentCoverage", "numPepsUnique", "scoreUnique", "species", "orfCategory", "accession_number",
                        "subgroupNum", "entry_name"]

                df = df.drop(columns=cols_to_drop)

                # Format table
                df = df.apply(pd.to_numeric)
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.index.name = "Patient_ID"

                self._data["proteomics"] = df

            elif file_name == "luad-v2.0-rnaseq-prot-uq-rpkm-log2-NArm-row-norm.gct.gz" or file_name == "luad-v3.1-rnaseq-prot-uq-rpkm-log2-NArm.gct.gz":
                df = pd.read_csv(file_path, sep="\t", skiprows=2, dtype=object)

                # Filter out metadata rows
                gene_filter = df['geneSymbol'] != 'na'
                df = df[gene_filter]

                df = df.set_index('geneSymbol')
                cols_to_drop = ['id', 'gene_id', 'gene_type', 'length']
                df = df.drop(columns = cols_to_drop)

                df = df.apply(pd.to_numeric)
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.index.name = "Patient_ID"
                df = df.sort_index()

                self._data["transcriptomics"] = df

            elif file_name == "luad-v3.0-rnaseq-circ-rna_parsed.tsv.gz" and self._version == "3.1.1":
                df = pd.read_csv(file_path, sep='\t', dtype={"spanning.reads": "int16"}, engine="c")
                df = df.reset_index() # More memory efficient to do this here, rather than save with a range index when we parse the file it originally
                df = df.pivot(index="Sample.ID", columns="geneID", values="spanning.reads")
                df.index.name = "Patient_ID"
                df = df.sort_index()
                self._data['circular_RNA'] = df

            elif file_name in ["luad-v2.0-rnaseq-circ-rna.csv.gz", "luad-v3.0-rnaseq-circ-rna.csv.gz"] and self._version in ["2.0", "3.1"]:
                df = pd.read_csv(file_path, sep=",")

                junct_3_split = df['junction.3'].str.split(':', n=2, expand=True)
                chrm = junct_3_split[0] # Get the chromosome
                three_prime = junct_3_split[1] # Get the nucleotide coordinate of the last base of the acceptor

                junct_5_split = df['junction.5'].str.split(':', n=2, expand=True)
                five_prime = junct_5_split[1] # Get the nucleotide coordinates of the first base of the donor

                # Now we need the gene name
                diff = df['gene.5'] != df['gene.3'] # Create a boolean filter where genes are different
                temp = df['gene.5'].where(diff, other="") # Replace the ones that are the same with an empty string
                gene_name = temp + '_' + df["gene.3"] # Concatentate the temp column(which only has the genes from gene.5 that are different) to gene.3

                # Put all those pieces of information together
                df = df.assign(geneID=chrm + '_' + five_prime + '_' + three_prime + '_' + gene_name)

                # Slice out the columns we want
                df = df[['geneID', 'spanning.reads', 'Sample.ID']]

                #There are about 3,000 duplicates in the file. Duplicate meaning that they have identical Sample IDs and identical geneID, but different spanning reads.
                # Marcin Cieslik said to drop the one with the lowest spanning read.
                df = df.sort_values(by='spanning.reads', ascending=False).drop_duplicates(['Sample.ID','geneID']).sort_index()

                df = df.pivot(index="Sample.ID", columns="geneID")['spanning.reads']
                df.index.name = "Patient_ID"
                df = df.sort_index()
                self._data['circular_RNA'] = df

            elif file_name == "luad-v2.0-sample-annotation.csv.gz" or file_name == "luad-v3.1-sample-annotation.csv.gz":
                df = pd.read_csv(file_path, sep=",")
                
                filter = df['QC.status'] == "QC.pass" # There are some samples that are internal references. IRs are used for scaling purposes, and don't belong to a single patient, so we want to drop them.
                df = df[filter]

                df = df.drop(columns="Participant") # Get rid of the "Participant" column becuase the same information is stored in Sample.ID  which is formatted the way we want.
                df = df.set_index("Sample.ID")
                df.index.name="Patient_ID"
                df = df.rename(columns={"Type":"Sample_Tumor_Normal"})
                df["Sample_Tumor_Normal"] = df["Sample_Tumor_Normal"].replace(to_replace="NAT", value="Normal")

                # Split the metadata into multiple dataframes

                # Make experimental_design dataframe
                if self._version == "2.0":
                    experimental_design_cols = ['Experiment', 'Channel', 'QC.status']
                elif self._version in ["3.1", "3.1.1"]:
                    experimental_design_cols = ["Experiment", "Channel", "Aliquot", "QC.status"]

                experimental_design_df = df[experimental_design_cols]
                experimental_design_df.insert(0, "Sample_Tumor_Normal", df["Sample_Tumor_Normal"].copy()) # This is useful in both tables
                df = df.drop(columns=experimental_design_cols)

                # Make derived_molecular dataframe
                if self._version == "2.0":
                    derived_molecular_cols = ['TP53.mutation', 'KRAS.mutation', 'STK11.mutation', 'EGFR.mutation', 'KEAP1.mutation', 'RB1.mutation',
                        'IL21R.mutation', 'EGFL6.mutation', 'LMO2.mutation', 'C10orf62.mutation', 'DKK3.mutation', 'BIRC6.mutation', 'TP53.mutation.status',
                        'KRAS.mutation.status', 'STK11.mutation.status', 'EGFR.mutation.status', 'KEAP1.mutation.status', 'RB1.mutation.status', 'IL21R.mutation.status',
                        'EGFL6.mutation.status', 'LMO2.mutation.status', 'C10orf62.mutation.status', 'DKK3.mutation.status', 'BIRC6.mutation.status',
                        'Mutation.Signature.Activity.W1.COSMIC5', 'Mutation.Signature.Activity.W2.COSMIC4', 'Mutation.Signature.Activity.W3.COSMIC2', 'fusion.EML4-ALK']

                elif self._version in ["3.1", "3.1.1"]:
                    derived_molecular_cols = ["Smoking.Score.WGS", "Smoking.Signature.Fraction.WGS", "Dominant.Signature.WGS.notSmoking.50perc",
                        "Dominant.Signature.Fraction.WGS.notSmoking", "DNP.GG.to.TT.or.CC.to.AA.Count.WGS", "NMF.consensus", "NMF.cluster.membership",
                        "mRNA.Expression.Subtype.TCGA", "mRNA.stemness.index", "CIMP.status", "Tumor.Purity.byESTIMATE.RNAseq", "TSNet Purity", "ESTIMATEScore",
                        "ESTIMATE ImmuneScore", "ESTIMATE StromalScore", "TP53.mutation", "KRAS.mutation", "STK11.mutation", "EGFR.mutation", "KEAP1.mutation",
                        "RB1.mutation", "IL21R.mutation", "EGFL6.mutation", "LMO2.mutation", "C10orf62.mutation", "DKK3.mutation", "BIRC6.mutation",
                        "BRAF.mutation", "ARAF.mutation", "ERBB2.mutation", "TP53.mutation.status", "KRAS.mutation.status", "STK11.mutation.status",
                        "EGFR.mutation.status", "KEAP1.mutation.status", "RB1.mutation.status", "IL21R.mutation.status", "EGFL6.mutation.status", "LMO2.mutation.status",
                        "C10orf62.mutation.status", "DKK3.mutation.status", "BIRC6.mutation.status", "BRAF.mutation.status", "ARAF.mutation.status",
                        "ERBB2.mutation.status", "Total.Mutation.Count.WGS", "Mutation.Count.ExcludingINDELs.WGS", "Total.DNP.Count.WGS", "Number.somatic.mutations",
                        "Mutation.Signature.Activity.W1.COSMIC5", "Mutation.Signature.Activity.W2.COSMIC4", "Mutation.Signature.Activity.W3.COSMIC2", "ALK.fusion",
                        "ROS1.fusion", "RET.fusion", "Putative.driver.mutation"]

                derived_molecular_df = df[derived_molecular_cols]
                df = df.drop(columns = derived_molecular_cols)

                self._data["clinical"] = df
                self._data['experimental_design'] = experimental_design_df
                self._data['derived_molecular'] = derived_molecular_df

            elif file_name == 'LUAD_followup_9_12.xlsx' and self._version in ["3.1", "3.1.1"]:
                df = pd.read_excel(file_path)

                # Replace redundant values for "not reported" with NaN
                nan_equivalents = ['Not Reported/ Unknown', 'Reported/ Unknown', 'Not Applicable',
                    'na', 'unknown', 'Not Performed', 'Unknown tumor status', 'Unknown',
                    'Unknown Tumor Status', 'Not specified']

                df = df.replace(nan_equivalents, np.nan)

                # Replace redundant values for cause of death
                disease_prog_equivalents = ['Progression of disease', 'Progression of disease ', 'Tumor', 'Disease progression',
                    'Progressive Disease', 'disease progression', 'disease progression ', 'main disease ']

                df['Cause of Death'] = df['Cause of Death'].replace(disease_prog_equivalents, 'Disease progression')

                # Rename, set, and sort by index
                df = df.rename(columns={"Case ID": "Patient_ID"})
                df = df.set_index("Patient_ID")
                df = df.sort_index()

                self._data["followup"] = df

        print(' ' * len(loading_msg), end='\r') # Erase the loading message
        formatting_msg = "Formatting dataframes..."
        print(formatting_msg, end='\r')

        # Get a union of all dataframes' indices, with duplicates removed
        # Exclude the followup dataframe because it has samples from a different cohort that aren't included anywhere else in the dataset
        master_index = unionize_indices(self._data, exclude="followup")

        # Use the master index to reindex the clinical dataframe, so the clinical dataframe has a record of every sample in the dataset. Rows that didn't exist before (such as the rows for normal samples) are filled with NaN.
        clinical = self._data["clinical"]
        clinical = clinical.reindex(master_index)
        self._data['clinical'] = clinical

        # Drop samples C3N.00545 and C3N.00545.N from the dataset. They were excluded due to poor sample quality (see data freeze README; excluded in data freeze 3.0)
        cases_to_drop = ["C3N.00545", "C3N.00545.N"]
        for name in self._data.keys(): # Loop over the keys so we can alter the values without any issues
            df = self._data[name]
            df = df.drop(index=cases_to_drop, errors="ignore")
            self._data[name] = df

        # Replace periods with hyphens in all Patient_IDs
        for name in self._data.keys(): # Loop over just the keys to avoid any issues that would come if we looped over the values while editing them
            df = self._data[name]

            # Make the index a column so we can edit it
            df.index.name = "Patient_ID"
            df = df.reset_index()
            
            # Replace all '.' with '-'
            df["Patient_ID"] = df["Patient_ID"].str.replace(r"\.", "-", regex=True) 
            df["Patient_ID"] = df["Patient_ID"].str.replace(r"-N$", ".N", regex=True) # If there's a "-N" at the end, it's part of the normal identifier, which we want to actually be ".N"
            
            # Set the index back to Patient_ID
            df = df.set_index("Patient_ID")

            self._data[name] = df

        # Call function from dataframe_tools.py to sort all tables first by sample status, and then by the index
        self._data = sort_all_rows(self._data)

        # Call function from dataframe_tools.py to standardize the names of the index and column axes
        self._data = standardize_axes_names(self._data)

        print(" " * len(formatting_msg), end='\r') # Erase the formatting message

    def how_to_cite(self):
        return super().how_to_cite(cancer_type='lung adenocarcinoma', pmid=32649874)